#!/usr/bin/perl 
#
# Czech Morphology Generator
# (C) 2000 Jan Hajic
#
# Parameters:
# 1 Dictionary name (complete dictionary: roots, endings)
# < Input file (iso-8859-2, <Gil>/<Git> marked, plus other SGML possible)
# > Output file (<Gel>/<Get>/<Gef>format, SGML preserved)
#

use FileHandle; # for server mode (autoflush)
use Socket; # for server mode
use Carp; # for server mode

# Version info:
$FMVERSION = "1.2.1"; # windows code handling added
$FMDATE = "2001/08/16";
#$FMVERSION = "1.2.0"; # correct handling of multiple form generation for a tag
#$FMDATE = "2000/11/11";
#$FMVERSION = "1.1.0"; # dynamic wildcard expansion based on lemma (speedup!)
#$FMDATE = "2000/09/02";
#$FMVERSION = "1.0.0"; # inital; compatible w/b4101 plus <Gei>, <Git>* for pos
#$FMDATE = "2000/08/28";

# PARAMETERS/OPTIONS: see sub SetOptions, make a copy and set it there

# GLOBAL variables:

# USER-SUPPLIED OR DEFAULT OPTIONS/PARAMETERS/ARGUMENTS:
# see sub SetOptiions

int $fServ; # 0 (default): stdin/stdout or $1/$2 normal run; 
            # 1 - daemon run: server for client requests; set in main
$szSERVER = "SERVER";
my $iPort;
$szNOSTDIO = "NOSTDIO";
$szMAIn = ""; # input file name if not stdin
$szMAOut = ""; # output file name if not stdout

my $cEndingCount = 0; # just stats - ending entries
my $cRootCount   = 0; # just stats - root entries
my $szDictIDAndVersionInfo;

int $fInputType; # input file type (SGML, plain)
int $fSGMLType; # if input is SGML, type (HTML, other SGML, CSTS)
int $fSGML;  $fSGML =  1; # input file type
int $fCSTS;  $fCSTS =  4; # input file type
int $fHTML;  $fHTML =  2; # input file type
int $fPlain; $fPlain = 3; # input file type

int $ISOLATIN2; $ISOLATIN2 = 1;
int $WINCP1250; $WINCP1250 = 2;
int $fInputCode; # current input coding: 
                 # only ISO Latin2 (1) and Win CP 1250 supported (2)

$szPar1 = ""; # optional command-line parameter (after dictionary name, $2)
$szPar2 = ""; # second optional command-line parameter ($3) (unused)

int $cEndingCountMax; # maximum number of characters in an ending in dictionary
                 # (set in GetDictionary)
# capitalization handling constants:
$iCapOrig = 0;
$iCapOnly = 1;
$iCapLower = 2;

# for case conversion, letter identification (GetGenToken and calls):
$szUpperAccented = "ݮϫǡ";
$szLowerAccented = "籶";
$reUpperAcc = "[$szUpperAccented]";
$reLowerAcc = "[$szLowerAccented]";

$opEnt = "&lt;";
$clEnt = "&gt;";
$entAmp = "&amp;";
$entSemi = "&semi;";
$entPercnt = "&percnt;";
$entSlash = "&slash;";
$entVbar = "&vbar;";

my $szFormResult; # global for generated result in SGML (Get/Gef pairs)
my $cPositLen = 0; # positional tag length
my $iWaitEndDefault = 4; # defualt wait at end
my $iWaitEnd = $iWaitEndDefault;

my $szForm1; # result golbal from MorphGenerateForm
my $szAddInfo; # additional info on lemma (pdgm, comment, synt, ...)

int $iRCTokenOK; 
    $iRCTokenOK = 1; # constant for next-token-available flag after GetToken

$STRICT = 1; # first pass when even conditional filters are in effect
$RELAXED = 2; # second pass if first produces nothing; conditional filters out

int $iPass; # for pass indication (STRICT, RELAXED)
$szFilterT = ""; # current tag filter, to be set depending on $iPass
$szFilterL = ""; # current (synt/sem/style) filter, to be set dep. on $iPass

# $hResultsCollected; # hash for results of a single wordform

sub wee2il2($) # convert string from Win Cp1250 to ISO Latin 2 #########
{
  my ($szIn) = @_;
  my $szOut;
  ($szOut = $szIn) =~ tr 
        [\232\212\235\215\236\216\234\214\237\217\204\224\205] 
        ["".] ;
  return $szOut;
} # of sub wee2il2 ######################################################

sub il22wee($) # convert string from ISO Latin 2 to Win Cp1250 ##########
{
  my ($szIn) = @_;
  my $szOut;
  ($szOut = $szIn) =~ tr 
        []
        [\232\212\235\215\236\216\234\214\237\217] ;
  return $szOut;
} # of sub il22wee #######################################################

sub PrintCoded($) ### print coded string in win CP 1250 if input was same
{
  my ($szIn) = @_;

  if ($fInputCode == $WINCP1250) {
    print il22wee($szIn);
  }
  else {
    print $szIn;
  }
  return 0;
} # of sub PrintCoded ####################################################

sub Ch2Entity($) # code parameter (single char) to SGML entity ###########
{
  # returns input char if conversion not needed
  my ($chIn) = @_;
  my $szEntity;
  if ($chIn eq "<") {
    $szEntity = $opEnt;
  }
  elsif ($chIn eq ">") {
    $szEntity = $clEnt;
  }
  elsif ($chIn eq "&") {
    $szEntity = $entAmp;
  }
# semicolon, after all, does not make problems
#  elsif ($chIn eq ";") {
#    $szEntity = $entSemi;
#  }
  elsif ($chIn eq "%") {
    $szEntity = $entPercnt;
  }
  elsif ($chIn eq "|") {
    $szEntity = $entVbar;
  }
  else {
    $szEntity = $chIn;
  }
  return $szEntity;
} # end of Ch2Entity (coding char to SGML entity name)
 
sub IsWS($) # checks input string for whitespace; treturns 1 if it is ####
{
  my ($szIn) = @_;
  my $iStart;
  my $ch;

  for ($iStart = 0; $iStart < length($szIn); $iStart++) {
    $ch = substr $szIn, $iStart, 1;
    if ("$ch" ne " " &&
        "$ch" ne "\t" &&
        "$ch" ne "\n" &&
        "$ch" ne "\r" &&     
        "$ch" ne "\f") {
      return 0;
    }
  } # end of string: no non-whitespace found
  return 1; # all whitespace
} # end of IsWS, whitespace ##############################################

sub StripWS($) # strip whitespace from beginning & end of string #########
{
  my ($szIn) = @_;
  my $iStart;
  my $iEnd;
  my $ch;
  my $szReturn;

  # find first non-whitespace from start of string:
  for ($iStart = 0; $iStart < length($szIn); $iStart++) {
    $ch = substr $szIn, $iStart, 1;
    if (!IsWS($ch)) { last; }
  }
  # find first non-whitespace from end but not earlier than $iStart:
  for ($iEnd = length($szIn) - 1; $iEnd >= $iStart; $iEnd--) {
    $ch = substr $szIn, $iEnd, 1;
    if (!IsWS($ch)) { last; }
  }
  if ($iEnd >= $iStart) { # something non-whitespace left there
    $szReturn = substr $szIn, $iStart, ($iEnd - $iStart + 1);
  }
  else { # empty string (nl and/or whitespace only)
    $szReturn = "";
  }
  return $szReturn;
} # end of StripWS #######################################################

sub StripWSandCode($) # strip whitespace, convert code to ISO if needed
{
  my ($szIn) = @_;
  my $szStripped;
  my $szCoded;

  $szStripped = StripWS($szIn);
  if ($fInputCode == $WINCP1250) {
    $szCoded = wee2il2($szStripped);
    return $szCoded;
  }
  else {
    return $szStripped;
  }
} # end of StripWSandCode ################################################

sub GetDictionaryGen($) # open and read dictionary ##########################
{
  my ($szDictFN) = @_; # get parameters
  my $iRC; # return code from this sub
  my $iGenPdgm = 0; # number of generated pdgms
  my $iPositLen; # current positional tag length

  $iRC = 0;
  $cPositLen = 0; # positional tag length (global)
  my %hFixedDone; # for dict tags: after expansion, remember what processed


  $cEndingCountMax = 0;
  $DICTIONARY = "<" . $szDictFN;
  print STDERR $szPN, ": Opening Dictionary ", $szDictFN, "\n";
  $iRCopen = open(DICTIONARY,$szDictFN);
  if (!defined $iRCopen) { # check if open ok
    return 1;
  }
  # read dictionary one line at a time:
  $szDictIDAndVersionInfo = "[Version info not available]";
  while (<DICTIONARY>) {
    chomp; # get rid of final whitespace incl. newline
    if ("$_" eq "") { next; }
    $chID = substr $_, 0, 1;
    if ($chID eq ";") {
      if ($_ =~ /[ ]@[ ]/) {
        ($szDummy,$szDictIDAndVersionInfo) = split /@[ ]/, $_, 2;
      }
      # otherwise ordinary comment; ignore it
    }
    elsif ($chID eq "E") { # "ending" line
      ($chIgn,$szPat,$fNeg,$fNej,$szEnding,$szTagc,$szTagp,$szJunk) =
	  split /\|/, $_; # splitting the input line into:
                         # paradigm name ($szPat)
                         # negation flag (0/1: $fNeg)
                         # superlative flag (0/1: $fNej)
                         # ending string ($szEnding)
                         # compact tag ($szTagc), positional tag ($szTagp)
      if (defined $szJunk ||
          !defined $szPat ||
          !defined $fNeg ||
          !defined $fNej ||
          !defined $szEnding ||
          !defined $szTagc ||
          !defined $szTagp) { # check if all fields present and nothing else
        print STDERR $szPN,
                   ": ERROR: ending data error: ",$DictionaryFileName,
                   ": line ", $_, "\n";
        return 3;
      }
      $iPositLen = length($szTagp);
      if ($cPositLen == 0) {
        $cPositLen = $iPositLen;
      }
      else {
        if ($iPositLen != $cPositLen) {
          print STDERR $szPN,
                   ": ERROR: positional tag length mismatch: ",
                    $DictionaryFileName,
                   ": line ", $_, "\n";
          return 33;
	}
      }

      # remember the longest ending's length:
      if (length($szEnding) > $cEndingCountMax) {
        $cEndingCountMax = length($szEnding);
      }
      # generate "rule(s)" based on neg/nej flags:
      if ($szEnding eq "0") {
	$szEnding = ""; # 0 in ending is empty string
      }
      for ($iNeg = 0; $iNeg < 2; $iNeg++) {
        if ($iNeg <= $fNeg) { # only if negation possible; no neg always
          for ($iNej = 0; $iNej < 2; $iNej++) {
            if ($iNej <= $fNej) { # only if sup possible; no sup always
              $szPref = "";
	      if ($iNej == 1) { $szPref .= "nej"; } # chg for SK, PL
	      if ($iNeg == 1) { $szPref .= "ne"; } # reverse for PL
              $szRuleGen = $szPref . "~" . $szEnding;  
              $szTagcFixed = Template2Tag($szTagc,$iNeg,$iNej);
              $szTagpFixed = Template2Tag($szTagp,$iNeg,$iNej);
              @rgEndData = ($szRuleGen,$iNeg,$iNej);
              push @{ $hPdgmTagc2Rules{$szPat}{$szTagcFixed} }, [ @rgEndData ];
              push @{ $hPdgmTagp2Rules{$szPat}{$szTagpFixed} }, [ @rgEndData ];
              $cEndingCount++;
#print STDERR "E,$szPat,$szRuleGen,$szTagcFixed,$szTagpFixed,$szJunk\n"; # dbg
            }
          }          
	}
      }
    }    
    elsif ($chID eq "R") { # "root" line
      ($chIgn,$szPat,$szRoot,$szLemma,
              $szTagc,$szTagp,$chAltPos,$szSem,$szStyle,$szComment,$szJunk) =
	  split /\|/, $_; # splitting the input line into:
                         # paradigm name ($szPat)
                         # root string ($szRoot)
                         # lemma idemtifier ($szLemma)
                         # compact tag ($szTagc), positional tag ($szTagp)
                         # alternate POS character ($chAltPos), obsolete
                         # terminology / semantics ($szSem)
                         # style etc. ($szStyle)
                         # human-readable comment
      if (defined $szJunk ||
          !defined $szPat ||
          !defined $szRoot ||
          !defined $szLemma ||
          !defined $szSem ||
          !defined $szStyle ||
          !defined $szComment ||
          !defined $chAltPos ||
          !defined $szTagc ||
          !defined $szTagp) { # check if all fields present and nothing else
        print STDERR $szPN,
                   ": ERROR: root data error: ",$DictionaryFileName,
                   ": line ", $_, "\n";
        return 4;
        $iPositLen = length($szTagp);
        if ($cPositLen == 0) {
          $cPositLen = $iPositLen;
        }
        else {
          if ($iPositLen != $cPositLen) {
            print STDERR $szPN,
                     ": ERROR: positional tag length mismatch: ",
                      $DictionaryFileName,
                     ": line ", $_, "\n";
            return 34;
	  }
        }
      }
      # save the root data (array of eight items):
      if ($szStyle !~ /x/) { # exclude "do-not-generate" entries 
        $szSingleLemma = $szLemma;
        $szLemmaRest = $szLemma;
        if ($szLemmaRest !~ /[`]/) {
          $szLemmaRest = ""; # no more than single pass needed -
                             # lemma is not multiple lemma
        }
        while ($szSingleLemma ne "") { # include full multiple lemma (`)
                                   # as well as parts
# if ($szLemma =~ /[`]/) { # dbg
#   print STDERR "Adding SingleLemma: $szSingleLemma\n"; } # dbg
          if ($szTagc eq "0" || $szTagp eq "0") { # regular entry
            @rgLemmaData = ($szRoot, $szPat, $szTagc, $szTagp, 
                           $chAltPos, $szSem, $szStyle,$szComment);
            push @{ $hLemma2PdgmRootEtc{$szSingleLemma} }, [ @rgLemmaData ];
          } # regular, hash saved
          # both tags should be 0 at the same time, but...
          if ($szTagc ne "0" || $szTagp ne "0") { # irregular
            undef %hFixedDone;
            for ($iNeg = 0; $iNeg < 2; $iNeg++) {
              for ($iNej = 0; $iNej < 2; $iNej++) {
                @rgLemmaData = ($szRoot, $szPat, # eight items 
                           $chAltPos, $szSem, $szStyle,$szComment,
                           $iNeg,$iNej);
                if ($szTagc ne "0") { # iregular entry
                  $szTagcFixed = Template2Tag($szTagc,$iNeg,$iNej);
                  if (!exists($hFixedDone{$szTagcFixed})) {
                    push @{ 
                      $hLemmaTagc2PdgmRootEtc{$szSingleLemma}{$szTagcFixed} }, 
                        [ @rgLemmaData ];
                    $hFixedDone{$szTagcFixed} = 1;
                  }
                }
                if ($szTagp ne "0") { # iregular entry
                  $szTagpFixed = Template2Tag($szTagp,$iNeg,$iNej);
                  if (!exists($hFixedDone{$szTagpFixed})) {
                    push @{ 
                      $hLemmaTagp2PdgmRootEtc{$szSingleLemma}{$szTagpFixed} }, 
                        [ @rgLemmaData ];
                           # print STDERR "R,$szPat,$szSingleLemma,$szRoot,",
                           #   "$szTagcFixed,$szTagpFixed,$szJunk\n"; # dbg
                    $hFixedDone{$szTagpFixed} = 1;
                  } # not yet in hash
                }
              } # inej loop
            } # ineg loop         
          } # at least one (normally both) irreg
          ($szSingleLemma,$szLemmaRest) = split /[`]/, $szLemmaRest, 2;
        } # end of while single lemma not empty
        $cRootCount++;
      } # end of not excluded becasue of Style x
    } # of R line ID   
    else { # wrong line ID: not a dictionary or a fomrat error
      print STDERR $szPN,
                   ": ERROR: dictionary format error: ",$DictionaryFileName,
                   ": line ", $_, "\n";
      return 2;
    }    
  } # end of dictionary reading loop: while (<DICTIONARY>)

  # print out the statistics:
  print STDERR $szPN,
               ": Dictionary ", $DictionaryFileName, " ",
               $szDictIDAndVersionInfo, "\n";
  print STDERR $szPN,
               ": Dictionary ", $DictionaryFileName, " read in; ",
               $cEndingCount, " ending, ", 
               $cRootCount, " root entries not excluded.\n";
  print STDERR $szPN, ": Max ending length: ", $cEndingCountMax, "\n";

  return 0;

} # end of GetDictionaryGen subroutine ######################################

sub InitializeGenReader($) # input file & reading routine initalization ######
{
  my $iRC = 0;

  my ($szInputData) = @_; # nonempty if input in parameter, 
                          # otherwise read STDIN
  $szLookAheadLine = ""; # initialize lookahead line
  $szLookAhead = ""; # lookahead string (rest of current line)
  $szCurrentLine = StripWS($szInputData); # current input line

  # get beginning of input: either file or parameter
  if ($szCurrentLine eq "") {
    while (<STDIN>) { # in fact, read once unless empty line(s) found at beg
      chomp;
      $szCurrentLine = StripWSandCode($_);
      # print STDERR "read raw: $_, $szCurrentLine\n"; # dbg
      if ("$szCurrentLine" ne "") { last; }
    }
    if ("$szCurrentLine" eq "") {
      return 1; # empty input file?
    }
  } # no data in parameter
  # print STDERR "read in: ", "|", $szCurrentLine, "|\n"; # dbg
  if (
     $szCurrentLine =~ /^[ \t\r\f\n]*<[cC][sS][tT][sS][ >]/ ||
     $szCurrentLine =~
 /^[ \t\r\f\n]*<[!][dD][oO][cC][tT][yY][pP][eE].*[^a-zA-Z0-9][cC][sS][tT][sS]/
      ) {
    $fInputType = $fCSTS;
    $fSGMLType = $fCSTS;
    $fSpacingInfo = 1; # must be set otherwise the output is screwed up (?gen)
  }
  elsif ($szCurrentLine =~ /^<[isD][ >]/) {
    $fInputType = $fCSTS;
    $fSGMLType = $fSGML;
  }
  elsif ($szCurrentLine =~ /^<Gi[lt][ >]/) {
    $fInputType = $fCSTS;
    $fSGMLType = $fSGML;
  }
  else {
    return 2; # bad input format
  }
  #print STDERR $szPN, ": Input file type: ", 
  #             $fInputType, " ", $fSGMLType, "\n"; # dbg

  $szLookAhead = $szCurrentLine;
  return $iRC;
} # end of InitializeGenReader ##############################################

sub InitializeGenWriter() # output file & writing initalization #############
{
  my $iRC = 0;

  if ($szSGMLDTD ne "" && ($fInputType != $fCSTS)) { # never do for the
             # time being when generating
    print "<$szSGMLDTD>\n"; # output heading: csts DTD used throughout
                            # if not changed in SetOptions
  }

  return $iRC;
} # end of InitializeGenWriter ##############################################

sub CloseGenReader() # input file & reading routine close ###################
{
  my $iRC = 0;

  return $iRC;
} # end of CloseGenReader ###################################################

sub CloseGenWriter() # output file & writing close ##########################
{
  my $iRC = 0;

  if ($szSGMLDTD ne "" && ($fInputType != $fCSTS)) {
    print "</$szSGMLDTD>\n"; # output heading close: csts DTD used throughout
                             # if not changed in SetOptions
  }

  return $iRC;
} # end of CloseGenWriter ###################################################

sub FormAttributes($) # analyze form for digits, capitalization etc. #####
{
                      # returns SGML attr value: "" (for lowercase)
                      # or mixed, upper, cap, num
  my $iRC = 0;
  my ($szFormIn) = @_;

  if ($szFormIn =~ /^[0-9]+([.,][0-9]+)*$/) {
    return "num";
  }
  if ($szFormIn =~ /[0-9]/) { # number somewhere
    return "mixed";
  }
  if ($szFormIn =~ /^[A-Z]$/ ||
      $szFormIn =~ /^($reUpperAcc)$/) { # single upper: cap
    return "cap";
  }
  if ($szFormIn =~ /^(([A-Z])|($reUpperAcc))+$/) { # all up > 2: upper
    return "upper";
  }
  if ($szFormIn =~ /^[A-Z]/ ||
      $szFormIn =~ /^($reUpperAcc)/) { # first uppercase, not
                                       # all: cap
    return "cap";
  }
  if ($szFormIn =~ /[A-Z]/ ||
      $szFormIn =~ /($reUpperAcc)/) { # upper in the middle
    return "mixed";
  }
  return ""; # default: all real letters, lowercase
} # end of FormAttributes ################################################

sub LetterToLower($) # lowercase one letter ##############################
{
  my ($ch) = @_;
  if ($ch =~ /^[A-Z]$/) { # ascii uppercase letter
    return (chr ((ord $ch) + 32)); 
  }
  if ((ord $ch) < 128) { # other ascii: return unchanged
    return $ch;
  }
  $i = index $szUpperAccented, $ch;
  if ($i < 0) { # not an uppercase accented letter: return unchanged
    return $ch;
  }
  else { # accented uppercase:
    return substr $szLowerAccented, $i, 1;
  }
} # of lowercase one letter ##############################################

sub LowercaseDown($) # lowercase all letters #############################
{
  my ($szIn) = @_;
  my $iLen; # length of input
  my $i; # current letter
  my $szResult;

  $szResult = "";
  $iLen = length $szIn;
  for ($i = 0; $i < $iLen; $i++) {
    $szResult .= LetterToLower(substr $szIn, $i, 1);
  }
  return $szResult;
} # end of LowercaseDown - lowercasing everything ########################

sub CapitalizeDown($) # lowercase all letters but the first ##############
{
  my ($szIn) = @_;
  my $szResult = substr $szIn, 0, 1; # keep first letter as is
  $szResult .= LowercaseDown(substr $szIn, 1); # lowercase the rest
  return $szResult; # return 'capitalized' version
} # end of CapitalizeDown - lowercasing everything but the first letter ##

sub GetGenToken() # The tokenizer: get one word form, plus surrounding junk #
{
  # returns the results in global variables:
  #    $szGapBef       junk before wordform, SGML-coded
  #    $szGapAft       junk after wordform, SGML-coded
  #    $szGotLemma     lemma from <Gil>
  #    $szGotTag       tag from <Git> (successively if multiple for one Gil)
  # returns 0 if eof, 1 if input continues (possibly)

  my $iRC = 1; # still something on input (assuming...)

  # initialize Output variables:
  $szGapBef = "";
  $szGapAft = "";
  $szForm = "";
  $szSGMLTagForm = "";
  $szAfterAnalysis = "";
  $szSpaceAfterForm = "";
  $fDotFollows = 0;
  $fHyphenFollows = 0;

  # print STDERR "GetGenToken() entered, lookahead (20 max): ", 
  #     (substr $szLookAhead, 0, 20), "\n"; # dbg

  SomethingToRead:
  while ($iRC == 1) { # still something to read
    while ("$szLookAhead" ne "") {
      if ($szLookAhead =~ /^<Gil>/) { # lemma
        ($szIgn,$szGotLemma) = split />/, $szLookAhead, 2;
        if ($szGotLemma =~ /</) {
          ($szGotLemma,$szLookAhead) = split /</, $szGotLemma, 2;
	  $szLookAhead = "<" . $szLookAhead;
	}
	else {
          $szLookAhead = "";
	}
        $szGotTag = "";
        return $iRC;
      }  
      elsif ($szLookAhead =~ /^<Git>/) { # tag
        ($szIgn,$szGotTag) = split />/, $szLookAhead, 2;
             # rint STDERR "got ign/tag; $szIgn,$szGotTag\n"; # dbg
        if ($szGotTag =~ /</) {
          ($szGotTag,$szLookAhead) = split /</, $szGotTag, 2;
	  $szLookAhead = "<" . $szLookAhead;
	}
	else {
          $szLookAhead = "";
	}
        $szGotLemma = "";
        return $iRC;
      }
      elsif ($szLookAhead =~ /^</) { # other SGML - copy only (whole line)
        ($szActLine, $szLookAhead) = split /[\n]/, $szLookAhead, 2; # split
                                     # at first newline
        #print STDERR "actline $szActLine, lookahead before split: |$szLookAhead|\n"; # dbg
        chomp $szActLine;
        # print STDERR "just copying: !$szActLine!\n"; # dbg
        $szGapBef .= $szActLine . "\n";
	$szActLine = ""; # $szLookAhead already set
      } 
      # ---------------- else: just junk ------------------------------------
      else { # convert junk to <i>-type tags too (like SGML tags)
        $chFirst = substr $szLookAhead, 0, 1; # first character of input
        $szT = sprintf "&#0x%02x;", ord $chFirst;
        $szGapBef .= $szSGMLJunk . $szT . "\n";
        $szLookAhead = substr $szLookAhead, 1, length($szLookAhead) - 1;
        if ($fIgnoreJunk == 1) { # no output of <i>-tags requested:
          $szGapBef = ""; # oh well, delete it...
        }
      }
    } # while lookahead not empty
    $szCurrentLine = "";
    if ($fServ == 1) { # server mode: do not read STDIN
      return 0; # lookahead empty, nothing more to do
    }
    while (<STDIN>) { # in fact, just read once unless empty line(s) found
      chomp;
      $szCurrentLine = StripWSandCode($_);
      if ("$szCurrentLine" eq "") { next; }
      # print STDERR "read in: ", $szCurrentLine, "\n"; # dbg
      last;
    } # end of reading stdin (while (<>))
    $szLookAhead = $szCurrentLine;
    if ("$szLookAhead" eq "") { 
      $iRC = 0; }
  } # of while reading (until form found)
  return $iRC;
} # end of tokenizer GetGenToken (getting wordform from input) ################

sub FilterTag($$) ###### filters tag out (based on SetOptions) #############
{
  my $iResult = 0; # default: do not filter out
  my ($szTagCurr, $szSuffList) = @_;
  # check if output filtering requested:
  if ($szSuffList ne "") { # check if tag should be
                                                # filtered out (per option)
    if ($fPositionalTags == 1) { # check last position
      $chTag = substr $szTagCurr, length($szTagCurr) - 1, 1; # last position
      if (index($szSuffList,$chTag) >= 0) { # found, del
        $iResult = 1; # ignore this result completely
      }
    } # end of positional tag check
    else { # checking compact tag:
      $iTagLen = length $szTagCurr;
      if ($iTagLen >= 2) { # tag suffix -<sym> possibly present
        if (substr $szTagCurr, $iTagLen - 2, 1 eq "-") { # yes...
          $chTag = substr $szTagCurr, $iTagLen - 1, 1; # last ch
          if (index($szSuffList, $chTag) >= 0) { # found
            $iResult = 1; # ignore this result completely
	  }
	} # end of tag suffix found in filter list
      } # end of tag long enough to check
    } # end of compact tag check for filtering out
  } # end of option to filter out based on tag set
  return $iResult;
} # end of FilterTag #######################################################

sub FilterString($$) ##### signals if string1 intersect with filter string #
{
  # signals if result should be ignored (rc == 1) based on presence of
  # any char from string1 in string 2;
  # ie. returns 1 if intersection non empty

  my $iResult = 0; # default: do not filter out
  my ($szStrCurr, $szSuffList) = @_; # $szSuffList is a string of chars
                                     # leading to filtering out of a result
                                     # if found in current string ($szStrCurr)
  my $i = 0; # for index
  my $cLen = 0; # length of option string containing letters to be filtered
  my $ch; # character from current string to be tested

  # check if output filtering requested:
  if ($szSuffList ne "" &&
      $szStrCurr ne "") { # ...and if anything to test in actual data
    $cLen = length($szStrCurr);
    for ($i = 0; $i < $cLen; $i++) {
      $ch = substr($szStrCurr, $i, 1);
      if (index($szSuffList,$ch) >= 0) { # found: filter it out
	$iResult = 1;
	last;
      } # end of found branch: set return result to "FILTER OUT" and exit loop
    } # end of loop over actual characters in current string
  } # end of if: filtering requested & makes sense
  return $iResult;
} # end of FilterString: signalling if filtering should occur ##############

sub Template2Tag($$$)
{
  my ($szTagTemplate, $fNegP, $fNejP) = @_;
  my $i;
  my $chTemplate;
  my $chTag;
  my $szTagResult = "";
  for ($i = 0; $i < length($szTagTemplate); $i++) {
    $chTemplate = substr $szTagTemplate, $i, 1;
    if ($chTemplate eq "@") { # negation; replace by A/N depending on presence
                              # of negative prefix in original form
      if ($fNegP == 1) { $chTag = "N"; }
      else { $chTag = "A"; }
    }
    elsif ($chTemplate eq "#") { # superlative template char; replace by 2/3
                                 # (comparative/superlative) depending on
                                 # the presence of the superlative prefix
      if ($fNejP == 1) { $chTag = "3"; }
      else { $chTag = "2"; }
    }
    else { # regular non-template char
      $chTag = $chTemplate;
    }
    $szTagResult .= $chTag; # add to resulting tag being built
  }
  return $szTagResult;  
} # end of Template2Tag

sub MorphAnalyzeBareForm($$$$$) # Analyze one wordform; case fixed, no pref #
{

  my $iRC = 0;
  # get parameters:
  my ($szFormIn, $fDot, $fHyph, $fNegStripped, $fNejStripped) = @_;

  my $iEnding; # current ending length
  my @rgRootDataCurr;
  my @rgrgEndCurr; # list of all data records for a current ending
  my @rgEndDataCurr;

#  print STDERR " - Analyzing: $szFormIn, flags dot $fDot hyph $fHyph,",
#                      " neg $fNegStripped, nej $fNejStripped \n"; # dbg

  $szFormInNP = $szFormIn;
  for ($iEnding = 0; $iEnding <= $cEndingCountMax; $iEnding++) {
    #
    # main analysis loop; successively segmenting input form into
    # root - ending pair, then checking for a paradigm match; if successful,
    # all morphological information is collected from both ending and root
    # and stored in $hResultsCollected (hash)
    #
    # start from the end, i.e. with root == form and ending == 0
    #
    if ($iEnding >= length($szFormInNP)) { last; } # finish if no root left
    if ($iEnding == 0) { # just started:
      $szEndingCurr = "0"; # for formal match with ending data
      $szRootCurr = $szFormInNP; # set root to full form
    }
    else { # second and successive passes go through here:
      $szEndingCurr = substr $szFormInNP, length($szFormInNP) - $iEnding, 
                                          $iEnding; # ending construction
      $szRootCurr = substr $szFormInNP, 0, length($szFormInNP) - $iEnding;
                    # root construction
    }
    # root + ending set now in $sz...Curr
    if (defined $hRootPdgm2LemmaEtc{$szRootCurr}) { # just for efficiency! 
      foreach $rgEndCurr (@{ $hEnding2PdgmEtc{$szEndingCurr} }) {
        # for every data entry in the current ending list,
        # find out if the current root matches:
        ($szPatCurr, $fNegCurr, $fNejCurr, $szTagcCurr, $szTagpCurr) = 
          @{ $rgEndCurr }; 
        # first filter out wrong analyses because of "external" conditions,
        # i.e. prefix mismatch, period/hyphen followup mismatch:
        if ($fNegCurr >= $fNegStripped) { # negation allowed or neg. prefix
                                          # not present
	  if ($fNejCurr >= $fNejStripped) { # sup. prefix allowed or
                                            # not present
            if ($szPatCurr eq "zkr" && $fDot == 0) { next; } # period req'd 
                                                             # but not present
            if ($szPatCurr eq "poml" && $fHyph == 0) { next; } # hyph req'd
            # ok, now really get the root data:
            if (exists $hRootPdgm2LemmaEtc{$szRootCurr}{$szPatCurr}) { 
              # match found!
              foreach $rgRootCurr 
                (@{$hRootPdgm2LemmaEtc{$szRootCurr}{$szPatCurr}}) {
                # for all roots which also match paradigm, get root info:
		($szLemmaDict, $szTagcDict, $szTagpDict, $chAltPosDict, 
                  $szSemDict, $szStyleDict, $szCommentDict) =
                  @{ $rgRootCurr};
                # first, build the tag part of result: ($szSGMLTagOut):
                if ($fPositionalTags == 1) {
    	          if ($szTagpDict ne "0") {
                    $szTagCurrTemplate = $szTagpDict;
		  }
		  else {
                    $szTagCurrTemplate = $szTagpCurr;
		  }
		} # positional tag ouput request selected by user
		else { # compact tag output
    	          if ($szTagcDict ne "0") {
                    $szTagCurrTemplate = $szTagcDict;
		  }
		  else {
                    $szTagCurrTemplate = $szTagcCurr;
		  }
		} # compact tag output selected by user
                # replace variables (@,#) for neg/superlative:
                $szTagCurr = Template2Tag($szTagCurrTemplate,
                                          $fNegStripped,
                                          $fNejStripped);
                # tag ready now
                if (FilterTag($szTagCurr,$szFilterT) > 0) {
                  next; # ignore current result completely
		}
                # prepare additional info (sem/style):
                $szAddInfo = "";
                if ($chAltPosDict ne "0") {
		  if (FilterString($chAltPosDict,$szFilterL) > 0) { next; }
	          for ($i = 0; $i < length($chAltPosDict); $i++) {
                    $chAdd = substr $chAltPosDict, $i, 1;
                    if ($chAdd ne substr $szTagCurr, 0, 1) {
                      $szAddInfo .= "_:" . $chAdd; # add if different than
                                                   # tag POS (comp/pos same!)
		    }
		  }
                } # end of adding syntax (alternate POS) info
                if ($szSemDict ne "0") {
		  if (FilterString($szSemDict,$szFilterL) > 0) { next; }
	          for ($i = 0; $i < length($szSemDict); $i++) {
                    $szAddInfo .= "_;" . substr $szSemDict, $i, 1;
		  }
                } # end of adding 'semantics' (terminology) info
                if ($szStyleDict ne "0") {
		  if (FilterString($szStyleDict,$szFilterL) > 0) { next; }
	          for ($i = 0; $i < length($szStyleDict); $i++) {
                    $szAddInfo .= "_," . substr $szStyleDict, $i, 1;
		  }
                } # end of adding style info
                if ($fCommentsOut == 1) {
                  if ($szCommentDict ne "0") {
                   $szAddInfo .= "_^"; # comment start separator 
	            for ($i = 0; $i < length($szCommentDict); $i++) {
                      $chAdd = substr $szCommentDict, $i, 1;
                      $szAddInfo .= Ch2Entity($chAdd); # eliminates potential 
                                                       # SGML problems
		    } # end of converting to entities where needed
	          } # end of comment contains something
                } # end of comment output
                $hResultsCollected{$szLemmaDict}{$szTagCurr} = $szAddInfo;
                                             # save result; replace prev.
                                             # if [pure] lemma & tag identical
	      } # end of collecting all dict data after a match had been ok'd
            } # end of root data exists for root/pdgm, i.e. match found
          } # end of superlative prefix match
        } # end of negative prefix match
      } # end of for each ending data record 
    } # root exists in dictionary (speedup test only)
    # else it makes no sense to check ending data - cannot be analyzed anyway

  } # of for (segmenting form to root + ending)

  return $iRC;
} # end of MorphAnalyzeBareForm ############################################

sub MorphAnalyzePrefixedForm($$$) # Analyze one wordform; case fixed #######
                                  # solve prefixing (ne-, nej-)
{
  my $iRC = 0;
  # get parameters:
  my ($szFormIn, $fDot, $fHyph) = @_;
  my $fNegPresent = 0;
  my $fNejPresent = 0;
  my $szFormWithoutPrefix;

  # first, call morph analysis with unchanged form:
  $iRCCall = MorphAnalyzeBareForm($szFormIn,$fDot,$fHyph,0,0); 
  if ($iRCCall != 0) { return 1; }
  if ($szFormIn =~ /^[nN][eE]/) { # form starts with ne-
    $szFormWithoutPrefix = substr $szFormIn, 2;
    $iRCCall = MorphAnalyzeBareForm($szFormWithoutPrefix,$fDot,$fHyph,1,0); 
    if ($iRCCall != 0) { return 2; }
  }
  if ($szFormIn =~ /^[nN][eE][jJ]/) { # form starts with nej-
    $szFormWithoutPrefix = substr $szFormIn, 3;
    $iRCCall = MorphAnalyzeBareForm($szFormWithoutPrefix,$fDot,$fHyph,0,1); 
    if ($iRCCall != 0) { return 3; }
  }
  if ($szFormIn =~ /^[nN][eE][jJ][nN][eE]/) { # form starts with nejne-
    $szFormWithoutPrefix = substr $szFormIn, 5;
    $iRCCall = MorphAnalyzeBareForm($szFormWithoutPrefix,$fDot,$fHyph,1,1); 
    if ($iRCCall != 0) { return 4; }
  }
  return 0;
} # end of MorphAnalyzePrefixedForm ########################################

sub MorphAnalyzeCapitalizedForm($$$$) # Analyze form; solve UC if any ######
        # solve capitalization:
        # - if original form in all uppercase or mixed: 
        #                 call analysis 3 times: UC/mix --> Cap --> LC
        # - if original form Capitalized only (or mixed w/no UC elsewhere):
        #                 call analysis twice: Cap --> LC
        # - if original lowercased:
        #                 LC only, one call only
        # Further (Cap/LC) analysis called regardless whether 
        #          success or not previously. (This rule might change
        #          unless the tagger takes care of it (preferred!))
{
  my $iRC = 0;
  # get parameters:
  my ($szFormIn, $szSGMLTag, $fDot, $fHyph) = @_;
  my $iCapitalizationLevel = 0; # 0 for unchanged, 1 for Capitalized only,
                                # 2 for all lowercase
  my $szFormCapOnly;
  my $szFormLower;

  # set global check/filter strings based on pass (1/2):
  if ($iPass == $STRICT) {
    $szFilterT = $szTagSuffixIgnore . $szTagSuffixIgnoreIfOtherOK;
    $szFilterL = $szLexIgnore . $szLexIgnoreIfOtherOK;
  }
  elsif ($iPass == $RELAXED) {
    $szFilterT = $szTagSuffixIgnore;
    $szFilterL = $szLexIgnore;
  }
  else { # internal fatal error - wrong $iPass number
    return 99;
  } # end of setting filtering strings

  undef %hResultsCollected; # initialize hash for result collection

  for ($iCapitalizationLevel = $iCapOrig; # loop three times: original caps,
       $iCapitalizationLevel <= $iCapLower; # then capitalized (1st letter)
       $iCapitalizationLevel++) {           # then all lowercase
    if ($iCapitalizationLevel == 0) {
      # always do this level; input form unchanged
      $iRCCall = MorphAnalyzePrefixedForm($szFormIn,$fDot,$fHyph);
                    # result: appended to in $szAnalyzedToken
      if ($iRCCall != 0) {
	return 1;
      }
    } # end of first pass code
    elsif ($iCapitalizationLevel == 1) { # capitalized (if different)
      $szFormCapOnly = CapitalizeDown($szFormIn);
      if ($szFormCapOnly eq $szFormIn) { # same as original form, do not proc.
        next;
      }
      $iRCCall = MorphAnalyzePrefixedForm($szFormCapOnly,$fDot,$fHyph);
                    # result: appended to in $szAnalyzedToken
      if ($iRCCall != 0) {
	return 2;
      }
    } # end of second pass code
    else { # all lowercase (if different from either of previous two)
      $szFormLower = LowercaseDown($szFormIn);
      if ($szFormLower eq $szFormIn ||
          $szFormLower eq $szCapOnly) { # already processed
        next;
      }
      $iRCCall = MorphAnalyzePrefixedForm($szFormLower,$fDot,$fHyph);
                    # result: appended to in $szAnalyzedToken
      if ($iRCCall != 0) {
	return 3;
      }
    } # end of third pass code
  } # of for loop for the capitalization levels
  return 0;
} # end of MorphAnalyzeCapitalizedForm #####################################

sub MorphAnalyzeForm($$$$) # Analyze one input form; solve capitalization ##
{
  my $iRC = 0;
  my $iRCCalled = 0;
  # get parameters:
  my ($szFormIn, $szSGMLTag, $fDot, $fHyph) = @_;

  undef %hResultsCollected; # initialize hash for result collection

  $iPass = $STRICT; # first pass: filter out everything incl. conditionals
                   #  -- see SetOptions
  $iRCCalled = MorphAnalyzeCapitalizedForm($szFormIn,$szSGMLTag,$fDot,$fHyph);
  if ($iRCCalled != 0) { # first pass failed fatally
    return 1;
  }
  if (%hResultsCollected <= 0) { # no result found after heavy 
                                 # (STRICT) filtering
    # try second pass, filter out only results unconditionally requested

    $iPass = $RELAXED;
    $iRCCalled = 
      MorphAnalyzeCapitalizedForm($szFormIn,$szSGMLTag,$fDot,$fHyph);
    if ($iRCCalled != 0) { # second pass failed fatally
      return 2;
    }
  }
  # output results now regardless if any success at any pass:
  # creating printable string of results:
  $szAnalyzedToken = ""; # prepare for concatenation of results
  if (%hResultsCollected > 0) { # yes, some analysis found (after filtering)
    foreach $szL (sort keys %hResultsCollected) { # for each lemma
      $szLemmaInfoOut = ""; # final lemma info (_....) to go out with $szL
      $szAnalyzedTokenTag = ""; # accumulated tags for current lemma in $szL
      $szAnalyzedToken .= $szSGMLLemmaOut . $szL; # Lemma output; 
                                                  # without info yet
      foreach $szT (sort keys %{ $hResultsCollected{$szL} }) {
        $szAnalyzedTokenTag .= $szSGMLTagOut . $szT; # add tag to list (string)
        $szLemmaInfo = $hResultsCollected{$szL}{$szT}; # get lemma info
        if (length($szLemmaInfo) > length($szLemmaInfoOut)) {
          if ($szLemmaInfoOut ne "") { # lemma info already set 
            print STDERR $szPN, ": incompatible lemma info ($szL/$szT)",
                                ": $szLemmaInfo vs. $szLemmaInfoOut", "\n";
	  } # end of printing error message
                # NB: this is kind of rough, if szLemmaInfoOut already set,
                # but probably works best; all
                # info should be same anyway for current lemma so we are only
                # making a soft-landing here for dictionary problems
          $szLemmaInfoOut = $szLemmaInfo; # replace old str (current is longer)
	} # end of storing lemma info anew
      } # end of loop over tags
      $szAnalyzedToken .= $szLemmaInfoOut . $szAnalyzedTokenTag;
    } # end of loop over lemmas
  } # end of: some result generated
  else { # "NOT FOUND": no solution found whatsoever (in the dictionary):
    if ($szSGMLTag eq $szSGMLPunctuation) { # "unknown" punctuation;
                                   # almost always handled here
      $szAnalyzedToken = $szSGMLLemmaOut . $szFormIn . $szSGMLTagOut;
      if ($fPositionalTags == 1) {
        $szAnalyzedToken .= "Z:-------------";
      }
      else { # compact tags requested 
        $szAnalyzedToken .= "ZIP";
      }
    } # end of handling: punctuation if not found
    elsif ($szSGMLTag =~ /^<($szSGMLForm)[ ]+(.*[ ])?num/) { 
                                              # number/digital expression
                                              # if not found in dict
      $szAnalyzedToken = $szSGMLLemmaOut . $szFormIn . $szSGMLTagOut;
      if ($fPositionalTags == 1) {
        $szAnalyzedToken .= "C=-------------";
      }
      else { # compact tags requested 
        $szAnalyzedToken .= "ZNUM";
      }
    } # end of handling: numerical/digital expressions
    else {
      $szAnalyzedToken = $szSGMLLemmaOut . $szFormIn . $szSGMLTagOut;
      if ($fPositionalTags == 1) {
        $szAnalyzedToken .= "X@-------------";
      }
      else { # compact tags requested 
        if (length($szFormIn) == 1) {
	  $szAnalyzedToken .= "NOMORPH1"; }
        else {
	  $szAnalyzedToken .= "NOMORPH"; }
      }
    } # end of handling: not a number, not punctuation
  } # end of final "NOT FOUND" handling section
  return 0;
} # end of MorphAnalyzeForm ################################################

sub MorphAnalyzeInput($) # morphologically analyze input (incl. tokenization)
{
  my $iRC = 0; # return code

  my ($szDataIn) = @_; # if nonempty, input in parameter, not STDIN

  $iRCCall = InitializeReader($szDataIn); # determine input file type,
                                          # set flags
  if ($iRCCall != 0) {
    print STDERR $szPN,
      ": Input initialization failed; check input file (stdin empty?)\n";
    return 1;
  }
  $iRCCall = InitializeWriter(); # write initial SGML to output 
  if ($iRCCall != 0) {
    print STDERR $szPN,
      ": Output initialization failed; check output file (stdout)\n";
    return 2;
  }

  ### main analyzer loop here: ###
  $fSomethingToRead = GetToken(); # read stuff into
                                  # $szGapBef, $szForm, $szGapAft,
                                  # $szSGMLTagForm,
                                  # $fDotFollows, $fHyphenFollows
  while ($fSomethingToRead != 0) {
              # print STDERR "Getoken returned:\n", "GB: ", $szGapBef, "\n", 
              #            "FM: ", $szForm, "\n", 
              #            "TG: ", $szSGMLTagForm, "\n", 
              #            "GA: ", $szGapAft, "\n",
              #            "DH: ", $fDotFollows,$fHyphenFollows,"\n"; # dbg
    $iRCCall = MorphAnalyzeForm($szForm,$szSGMLTagForm,
                                $fDotFollows,$fHyphenFollows);
                    # result: in $szAnalyzedToken, incl. $szForm and
                    # proper SGML tags
    if ($iRCCall != 0) {
      print STDERR $szPN,
                   ": ERROR analyzing ", $szForm, "; input line ", $iLineInput,
                   "; exiting\n";
      print "<<<ERROR>>>\n";
      $iRC = 6;
      last; # exit the loop;
    }
    else { # everything went ok
      if ("$szGapBef" ne "") {
	print $szGapBef; # newline already in there
      }
      if ($fNoTokenSGML == 0) { # suppresion of SGML tags in from of 
                                # tokens not req'd in SetOptions
        print $szSGMLTagForm;
      }
      print $szForm;
      $szAfterAnalysisBeforeResult = "";
      $szAfterAnalysisAfterResult = "";
      chomp $szAfterAnalysis;
      chomp $szAfterAnalysis;
      if ($szAfterAnalysis ne "") { # split to keep original order:
        ($szAfterAnalysisBeforeResult,$szAfterAnalysisAfterResult) 
          = split /<>/, $szAfterAnalysis, 2;
        $szAfterAnalysisAfterResult =~ s/<>//g; # get rid of remaining <>s 
      }
      if ($szAfterAnalysisBeforeResult ne "") {
        print $szAfterAnalysisBeforeResult;
      }
      print $szAnalyzedToken;
      if ($szAfterAnalysisAfterResult ne "") {
        print $szAfterAnalysisAfterResult;
      }
      print "\n";
      if ($szSpaceAfterForm ne "") { # also empty if fSpacingInfo disabled
        print $szSpaceAfterForm, "\n";
      }
      if ("$szGapAft" ne "") {
	print $szGapAft; # newline already in there
      }
    }

    $fSomethingToRead = GetToken(); # read next word; if everything ok, 
                                   # read stuff is in: $szGapBef, 
                                   # $szForm, $szSGMLTagForm, $szGapAft,
                                   # $fDotFollows, $fHyphenFollows
  } # end of the main word-reading loop (through GetToken())

  # still have to check for any gap-like material at the end of file:
  if ("$szGapBef" ne "") {
    print $szGapBef; # newline already in there
  }
  if ("$szGapAft" ne "") {
    print $szGapAft; # newline already in there
  }

  $iRCCall = CloseWriter(); # close output file (write final SGML tag(s))
  if ($iRCCall != 0) {
    print STDERR $szPN,
      ": Output file close failed; check output file (stdout)\n";
    return 3;
  }
  $iRCCall = CloseReader(); # close input file
  if ($iRCCall != 0) {
    print STDERR $szPN,
      ": Input file close failed;\n";
    return 4;
  }

  return $iRC;
} # end of MorphAnalyzeInput #################################################

sub ApplyRule($$) ######################### ApplyRule ########################
{
  # Applies rule $szRule to $szRoot
  # result returned
  my $iRC = 0;
  my ($szRoot,$szRule) = @_;
  my $szFormRes = ""; # result string

  my @rgRule = split //, $szRule;
  my $cRule = scalar(@rgRule);
  #       print STDERR "Rule $szRule has $cRule elements.\n"; # dbg dbg
  my $i;
  for ($i = 0; $i < $cRule; $i++) {
    $ch = $rgRule[$i];
    if ($ch eq "~") { # command to copy root here
      $szFormRes .= $szRoot;
    } # of root-copy "command"
    else { # just copy char
      $szFormRes .= $ch;
    } # of just copy char
  } # of for $i
  # print STDERR "Form generated: $szFormRes\n"; # dbg
  undef @rgRule; # should do by itself
  return $szFormRes;
} # end of ApplyRule ###################### ApplyRule ########################

sub MorphGenerateForm($$$) ####### generate word form(s) from 1 tag #######
{
  # innermost generating proedure - uses dictionary
  # result to ${3}
  my $iRC = 0;
  my ($szGIL,$szGIT,$phResTF) = @_;
  $szForm1 = ""; # result string
  my $iNeg;
  my $iNej;
  my $fLemmaFoundSomewhere = 0;
  my $iResults = 0;
  my %hRootsRules;
  my $szRR;
  my @rgRes;

  $szRuleGen = "";
  undef %hRootsRules;

# print STDERR "MGF called with $szGIL/$szGIT\n"; # dbg
  if ($fPositionalTags == 1) {
    $hLT2PRef = \%hLemmaTagp2PdgmRootEtc;
    $hPT2RulesRef = \%hPdgmTagp2Rules;
  }
  else {
    $hLT2PRef = \%hLemmaTagc2PdgmRootEtc;
    $hPT2RulesRef = \%hPdgmTagc2Rules;
  }
  if (exists($$hLT2PRef{$szGIL})) {
    $fLemmaFoundSomewhere = 1;
  }
  if (exists($$hLT2PRef{$szGIL}{$szGIT})) {
    LemmaLoop:
    foreach $rgIRLemmaCurr (@{$$hLT2PRef{$szGIL}{$szGIT}}) {
      ($szRoot, $szPat,
           $chAltPos, $szSem, $szStyle,$szComment,
           $iNegDict,$iNejDict) = @{ $rgIRLemmaCurr };
#           print STDERR "Retr irreg $szRoot,$szPat,$iNegDict,$iNejDict",
#                   " for $szGIL/$szGIT\n"; # dbg
      $szRuleGen = "";
#              print STDERR "Gen irreg posit MATCH? $szRoot",
#                 " for $szGIT, ",
#                 "pdgm $szPat "; # dbg
      # get any (first) rule from endings:
      FindRule:
      foreach $szTagKey (keys %{ $$hPT2RulesRef{$szPat} }) {
#            print STDERR "tagkey from endings: $szTagKey\n"; # dbg
        foreach $rgRule 
                (@{ $$hPT2RulesRef{$szPat}{$szTagKey} }) {
          ($szRuleGen,$iRuleNeg,$iRuleNej) = @{ $rgRule }; # take 1st
#              print STDERR "neg/nej endings: $iRuleNeg,$iRuleNej\n"; # dbg
          if ($iNegDict == $iRuleNeg &&
              $iNejDict == $iRuleNej) { # rule attributes for neg, nej fit
#    print STDERR "Gen irreg from $szGIL, for $szGIT, root $szRoot, rule $szRuleGen\n"; # dbg
            $szRR = $szRoot . "=@=" . $szRuleGen;
            @{ $hRootsRules{$szRR} } = 
                    ($chAltPos, $szSem, $szStyle, $szComment, $szPat);
            ### last FindRule;
          } # of getting rule
          else {
	    $szRuleGen = "";
          }
        } # tag (key) matches input
      } # of getting key
      ### if ($szRuleGen ne "") {
      ###  last LemmaLoop; # rule ok, go generate form
      ### }
    } # foreach lemma record
  } # of irreg entry exists 
  ### if ($szRuleGen eq "") { # irreg does not exist or no match
    if (exists($hLemma2PdgmRootEtc{$szGIL})) {
      $fLemmaFoundSomewhere = 1;
      Lemmas:
      foreach $rgLemmaCurr (@{$hLemma2PdgmRootEtc{$szGIL}}) {
        ($szRoot, $szPat, $szTagc, $szTagp, 
             $chAltPos, $szSem, $szStyle,$szComment) = @{ $rgLemmaCurr };
        if ($fPositionalTags == 1) {
	  $szTagDict = $szTagp;
        }
        else {
	  $szTagDict = $szTagc;
        }
        $szRuleGen = "";
        if ($szTagDict eq "0") { # use paradigms with endings etc. - 
                                 # regular case
#    print STDERR "Gen regular from $szGIL, for $szGIT, root $szRoot, pdgm $szPat", "\n"; # dbg
          foreach $rgRule 
              (@{ $$hPT2RulesRef{$szPat}{$szGIT} }) {
            ($szRuleGen,$iNeg,$iNej) = @{ $rgRule }; # take 1st
                          # (should be one only for reg. patterns anyway)
#    print STDERR "Tag $szGIT, reg, will use rule $szRuleGen/$szRoot\n"; # dbg
            $szRR = $szRoot . "=@=" . $szRuleGen;
            @{ $hRootsRules{$szRR} } = 
                    ($chAltPos, $szSem, $szStyle, $szComment, $szPat);
            #### last Lemmas; # should be only one rule anyway
          }
        } # end of Tag 0, ie. regular dict entry
      } # for each lemma in reg part of dict
    } # exists lemma in reg part of dict
  ### } # end of if: no irreg rule found, searching for reg entry
  # have rule(s) in $hRootsRules (maybe)
  foreach $szRR (keys %hRootsRules) {
    ($szRoot,$szRuleGen) = split /=@=/, $szRR, 2;
    ($chAltPos, $szSem, $szStyle, $szComment, $szPat) = 
	@{ $hRootsRules{$szRR} };
    ### if ($szRuleGen ne "") { # rule valid
# print STDERR "common end: applying rule $szRuleGen/$szRoot\n"; # dbg
# print STDERR " ap/sem/st/comm: $chAltPos,$szSem,$szStyle,$szComment|\n"; # dbg
    $szForm1 = ApplyRule($szRoot,$szRuleGen);
    if ($szForm1 ne "") {
      # prepare additional info (sem/style):
      $szAddInfo = "";
      if ($chAltPos ne "0") {
        if (FilterString($chAltPos,$szFilterL) > 0) { next; }
        for ($i = 0; $i < length($chAltPos); $i++) {
          $chAdd = substr $chAltPos, $i, 1;
          if ($chAdd ne substr $szTagCurr, 0, 1) {
            $szAddInfo .= "_:" . $chAdd; # add if different than
                                         # tag POS (comp/pos same!)
	  }
	}
      } # end of adding syntax (alternate POS) info
      if ($szSem ne "0") {
	# if (FilterString($szSem,$szFilterL) > 0) { next; }
        for ($i = 0; $i < length($szSem); $i++) {
          $szAddInfo .= "_;" . substr $szSem, $i, 1;
	}
      } # end of adding 'semantics' (terminology) info
      if ($szStyle ne "0") {
	# if (FilterString($szStyle,$szFilterL) > 0) { next; }
	for ($i = 0; $i < length($szStyle); $i++) {
          $szAddInfo .= "_," . substr $szStyle, $i, 1;
	}
      } # end of adding style info
      if ($fCommentsOut == 1) {
        if ($szComment ne "0") {
          $szAddInfo .= "_^"; # comment start separator 
	  for ($i = 0; $i < length($szComment); $i++) {
            $chAdd = substr $szComment, $i, 1;
            $szAddInfo .= Ch2Entity($chAdd); # eliminates potential 
                                             # SGML problems
	  } # end of converting to entities where needed
        } # end of comment contains something
      } # end of comment output
      if ($fPdgmOut == 1) {
	$szAddInfo .= "_@";
        $szAddInfo .= $szPat; # suppose szPat has no weird chars
      }
      # add result here:
#   print STDERR "adding form/info: $szForm1/$szAddInfo\n"; # dbg
      @rgRes = ($szForm1,$szAddInfo); 
      push @{ $$phResTF{$szGIT} }, [ @rgRes ];
      $iResults++;
    } # of if something generated for tag
    else {
              # else error in rule; do not output (replace if prev) result
#      print STDERR "Rule N.A.: $szRuleGen at $szGIT\n"; # dbg
    }
  } # of loop over roots/rules in hRootsRules
  ### } # of rule ok, processed
  ### else { # rule empty
        # print STDERR "Tag or other mismatch - not processed\n"; # dbg
  ### }
  if ($fLemmaFoundSomewhere != 1) {
    $iRC = -1; # lemma not found - no result possible
  }
  return $iRC;
} # end of MorphGenerateForm #################################################

sub TagsForLemma($) ###########################  generate possible tags ######
{
  # generates a list of all tags for a lemma, regardless if reg or irreg
  # returns a list of them, or empty list if lemma not found
  my @rgTagList;
  my @rgRegTags;
  my ($szLemma) = @_;
  my $szRoot;
  my $szPat;
  my $szTagc;
  my $szTagp;
  my $chAltPos;
  my $szSem;
  my $szStyle;
  my $szComment;
  my $szTagDict;

  undef @rgTagList;
  undef @rgRegTags;
# print "TagsForLemma called with $szLemma\n"; # dbg
  if ($fPositionalTags == 1) {
    $hLT2PRef = \%hLemmaTagp2PdgmRootEtc;
    $hPT2RulesRef = \%hPdgmTagp2Rules;
  }
  else {
    $hLT2PRef = \%hLemmaTagc2PdgmRootEtc;
    $hPT2RulesRef = \%hPdgmTagc2Rules;
  }
  # first, get tags from irregular (root record) entries if any:
  if (exists($$hLT2PRef{$szLemma})) {
    @rgTagList = keys %{ $$hLT2PRef{$szLemma} }; # possibly only 0 if no irreg
  }
# print "tags irrreg: @rgTagList\n"; # dbg
  # then, go to endings to retrieve all tags from a paradigm (for every rec!)
  if (exists($hLemma2PdgmRootEtc{$szLemma})) {
    # for all entries:
    Lemmas:
    foreach $rgLemmaCurr (@{$hLemma2PdgmRootEtc{$szLemma}}) {
      ($szRoot, $szPat, $szTagc, $szTagp, 
           $chAltPos, $szSem, $szStyle,$szComment) = @{ $rgLemmaCurr };
        if ($fPositionalTags == 1) {
	  $szTagDict = $szTagp;
        }
        else {
	  $szTagDict = $szTagc;
        }
      if ($szTagDict ne "0") {
        print STDERR "Tag $szTagDict not 0 for regular lemma $szLemma??\n";
        exit 433; # internal error!!
      }
      if (exists($$hPT2RulesRef{$szPat})) {
        @rgRegTags = keys %{ $$hPT2RulesRef{$szPat} };
# print "reg tags for $szPat: @rgRegTags\n"; # dbg
        push @rgTagList, @rgRegTags; # add to list of tags
# print "all tags: @rgTagList\n"; # dbg
      } # end of getting tags for a singl paradigm
    } # end of foreach lemma record in reg hash 
  } # end of lemma is regular
  undef @rgRegTags;
  return @rgTagList;
} # end of TagsForLemma #######################################################

sub MorphGenerateForms($$) ################# generate word form(s) ############
{
  my $iRC = 0;
  my ($szGIL,$szGIT) = @_;
  my %hExpandedTags;
  my $szTag;
  my %hResTag2Form; # accumulates results
  undef %hResTag2Form; # used locally before conversion to result string
  my @rgTags;
  my $szGITFullLen;
  my $iLen;

#  print STDERR "called MorphGenerateForms with $szGIL/$szGIT\n"; # dbg

  $fRegPatternNontrivial = 1;
  undef %hExpandedTags;
  if ($fPositionalTags == 1) {
    if ($szGIT eq "*") {
      $szRETagPattern = "." x $cPositLen;
    }
    else {
      $iLen = length($szGIT);
      $szGITFullLen = $szGIT;
      if (substr($szGIT,$iLen-1,1) eq "*") { # * at end of posit
                     # is allowed and means the same as for compact tags
        $szGITFullLen = substr($szGIT,0,$iLen-1) . 
                                 ("." x ($cPositLen - $iLen + 1));
      }
      if (length($szGITFullLen) != $cPositLen) { 
        print STDERR 
             "Wrong tag pattern length: $szGIT (should be $cPositLen)\n";
        return 101; # positional tag of wrong length
      }
      if ($szGITFullLen !~ /[.]/) { # no dot - not a real RE
        $fRegPatternNontrivial = 0;
      }
      ($szRETagPattern = $szGITFullLen) =~ 
                                 s/([\?\*\\\|\(\)\[\{\^\$\+])/\\$1/g; 
                                 # protect weird chars
                                 # except for . here
    }
  }
  else { # compact tags
    if ($szGIT !~ /[.?*]/) { # no dot/?/* - not a real RE
      $fRegPatternNontrivial = 0;
    }
    ($szRETagPattern = $szGIT) =~ s/([\\\|\(\)\[\{\^\$\+])/\\$1/g; 
                                 # protect weird chars
                                 # except for . ? * here
    if ($szRETagPattern =~ /\*$/ &&
        $szRETagPattern !~ /\.\*$/) { # change * into .* to match anything
      $szRETagPattern =~ s/\*$/.*/;
    }
  }

  if ($fRegPatternNontrivial == 1) {
    @rgTags = TagsForLemma($szGIL); # get all tags for a lemma
                                    # in fact needed only if underspec. templ.

#          print STDERR "Pattern used: $szRETagPattern\n"; # dbg
#          print STDERR "Taglist: @rgTags\n"; # dbg
    $szREComp = eval { qr/$szRETagPattern/ } || do {
      print STDERR "Pattern $szGIT for tags not a valid (shell-like) RE.\n";
      return 102;
    };
    foreach $szTag (@rgTags) {
      if ($szTag =~ /^$szREComp$/) {
        $hExpandedTags{$szTag} = 1;
# print "hashing to exp: $szTag\n"; # dbg
      }
    } # for expanding tag loop
    undef @rgTags;
  } # end of nontrivial RE in pattern
  else { # trivial RE i.e. "fixed" tag on input:
    $hExpandedTags{$szGIT} = 1;
  }
  # now all tags requested in %hExpandedTags
  $szFormResult = ""; # global for composite result
  foreach $szTag (%hExpandedTags) { # generate exactly one form
                # if choice genereateform will select !one
    $iRC = MorphGenerateForm($szGIL,$szTag,\%hResTag2Form); 
                        # result(s) to hResTag2Form
    if ($iRC > 0) {
      return 1; # cannot generate one form (fatal errors only!)
    }
    elsif ($iRC < 0) { # lemma not found in dict at all
      $iRC = 0;
      last; # no chance to get any result even for other tags
    }
  } # generating for each tag from pre-constructed list
  undef %hExpandedTags;
  foreach $szTag (sort keys %hResTag2Form) {
    @rgAllRes = @{ $hResTag2Form{$szTag} };
    for $irgRes ( 0 .. $#rgAllRes ) {
      $prgRes = $rgAllRes[$irgRes];
      ($szForm1,$szAddInfo) = @{ $prgRes };
      if ($fAddInfoOut == 1 && $szAddInfo ne "") {
        $szFormResult .= "$szSGMLGInfoOut" . $szAddInfo;
      }
      $szFormResult .= "$szSGMLGTagOut" . $szTag . "$szSGMLGFormOut" . $szForm1;
    }  
  } # of for each tag in results
  undef %hResTag2Form; # should do it by itself, but...
  return $iRC;
} # end of MorphGenerateForms

sub MorphGenerate($) # generate output forms from input file (<Gil>/<Git>
                     # format)
{
  my $iRC = 0; # return code

  my ($szDataIn) = @_; # if nonempty, input in parameter, not STDIN

  my $szGapFront = "";
  my $szGapBack = "";
  my $fPrevFormsOut;
  $iRCCall = InitializeGenReader($szDataIn); # check input file type,
                                          # set flags
  if ($iRCCall != 0) {
    print STDERR $szPN,
      ": Input initialization failed; check input file (stdin empty?)\n";
    return 1;
  }
  $iRCCall = InitializeGenWriter(); # write initial SGML to output 
  if ($iRCCall != 0) {
    print STDERR $szPN,
      ": Output initialization failed; check output file (stdout)\n";
    return 2;
  }

  ### main generating loop here: ###
  $fSomethingToRead = GetGenToken(); # read stuff into
                                # $szGapBef, $szGapAft,
                                # $szGotLemma, $szGotTag
  $fPrevFormsOut = 1; # signal if previous lemma/tag --> form output
  while ($fSomethingToRead != 0) {
            # print STDERR "GetGenToken returned:\n", "GB: ", $szGapBef, "\n", 
            #             "GL: ", $szGotLemma, "\n", 
            #             "GT: ", $szGotTag, "\n", 
            #             "GA: ", $szGapAft, "\n"
            #             ; # dbg

    # check what's returned:
    if ($szGotLemma ne "") { # got lemma
      if ($fPrevFormsOut == 0) { # must print out results for previous 
                                   # Gil/Git request (not found)
        PrintCoded($szGapFront);
        PrintCoded($szSGMLGLemma . $szGILemma . "\n"); 
        PrintCoded($szGapBack);
        $szGapFront = ""; # !maybe redundant if following assgn = only not .=
        $szGapBack = ""; # !maybe redundant if following assgn = only not .=
      }
      $szGapFront .= $szGapBef; 
      $szGapBack .= $szGapAft;
      $szGILemma = $szGotLemma;
      $fPrevFormsOut = 0;
      $szGITag = "";
    } # end of lemma returned from GetGenToken
    elsif ($szGotTag ne "") { # got tag!
      $szGapFront .= $szGapBef; # should be all empty here, but...
      $szGapBack .= $szGapAft;
      $szGITag = $szGotTag;
      if ($fPrevFormsOut == 0) { # othws ignore because previous tag [template]
                      # already generated some results, printed out!
        # call the word form generation here:
        $szFormResult = "";
        $iRC = MorphGenerateForms($szGILemma,$szGITag); 
                                          # result in $szFormResult
                                          # global string
        if ($iRC != 0) {
          print STDERR "Generating form(s) from Gil/Git ",
    	  "$szGILemma/$szGITag failed.\n";
          return 7;
        }
        if ($szFormResult ne "") { # result generated successfully!
          PrintCoded($szGapFront);
          PrintCoded($szSGMLGLemma . $szGILemma . $szFormResult . "\n"); 
          PrintCoded($szGapBack);
          $szGapFront = "";
          $szGapBack = "";
          $fPrevFormsOut = 1;
        }
      } # of if $fPrevFormsOut
    } # got tag from GetGenToken
    else {
      print STDERR "Gen: Internal error: GetGenToken returned wrong values\n";
      return 6; 
    }
    $fSomethingToRead = GetGenToken(); # read next word; if everything ok, 
                                   # read stuff is in: $szGapBef, etc.
  } # end of the main word-reading loop (through GetGenToken())

  # still have to check for any gap-like material at the end of file:
  if ($fPrevFormsOut == 0) { # not found
    PrintCoded($szGapFront);
    PrintCoded($szSGMLGLemma . $szGILemma . "\n"); 
    PrintCoded($szGapBack);
    $szGapFront = "";
    $szGapBack = "";
    $fPrevFormsOut = 1;
  }

  $iRCCall = CloseGenWriter(); # close output file (write final SGML tag(s))
  if ($iRCCall != 0) {
    print STDERR $szPN,
      ": Output file close failed; check output file (stdout)\n";
    return 3;
  }
  $iRCCall = CloseGenReader(); # close input file
  if ($iRCCall != 0) {
    print STDERR $szPN,
      ": Input file close failed;\n";
    return 4;
  }

  return $iRC;
} # end of MorphGenerate #################################################

sub SetOptionsDemoGen() ########## Set demo (interactive) mode ###############
{
  $szSGMLDTD = ""; # no heading
  $szSGMLGLemma = ""; # output lemma (from input, as found in dict)
  $szSGMLGFormOut = " "; # "info" tag about dict entry used 
  $szSGMLGTagOut = "\n\t\t"; # POS-tag used to generate following <Gef>
  $szSGMLGInfoOut = "\n\t"; # output form (generated) SGML tag
  $fIgnoreJunk = 1; # 1: ignore SGML code and other junk on input
  $fAddInfoOut = 1; # 1: output the additinal info in <Gei>
  $fCommentsOut = 1; # 1: output the human-readable comments (stuff in paren)
  $fPdgmOut = 1; # 1: output the paradigm name
  # keep SGML tags intact (irrelevant anyway)
} # end of SetOptionsDemo ####################################################

sub SetOptionsBatchGen() ########## Set batch mode (basic) ###################
{
  $szSGMLDTD = "csts"; # SGML tag name for output format DTD
                       # printed out as the first & last line
                       # !! Change NOT recommended.
                       # default: "csts"
                       # use "" for no DTD start/end tags at all
                       # NB: must change if you change any of the following
                       # SGML tags!
  $szSGMLGLemma = "<Gel>"; # output lemma (from input, as found in dict)
                             # Default: "<Gel>"
                             # Changing to non-SGML STRONGLY discouraged
                             # (only space or ctrl chars or weird
                             # characters > 128 which are not letters
                             # would work!!)
  $szSGMLGInfoOut = "<Gei>"; # "info" tag about dict entry used 
                             # to generate following form
                             # Default: "<Gei>"
                             # Changing to non-SGML STRONGLY discouraged
                             # (only space or ctrl chars or weird
                             # characters > 128 which are not letters
                             # would work!!)
  $szSGMLGTagOut = "<Get>"; # POS-tag used to generate following <Gef>
                           # Default: "<Get>"
                           # Changing to non-SGML STRONGLY discouraged
                           # (only space or ctrl chars or weird
                           # characters > 128 which are not letters
                           # would work!!)
  $szSGMLGFormOut = "<Gef>"; # output form (generated) SGML tag
                             # Default: "<Gef>"
                             # Changing to non-SGML STRONGLY discouraged
                             # (only space or ctrl chars or weird
                             # characters > 128 which are not letters
                             # would work!!)
  $fIgnoreJunk = 0; # 1: ignore SGML code and other junk on input
                    # default: 0: output everything (SGML-safely coded)
                    # in <i> tags, or whatever $szSGMLJunk says
  $fPositionalTags = 1; # 1: output in positional tag system
                        # 0: compact
                        # Default: 1 (positional tags)
  $fCommentsOut = 0; # 1: output the human-readable comments (stuff in paren)
  $fAddInfoOut = 0; # 1: output the additinal info in <Gei>
  $fPdgmOut = 0; # 1: output the paradigm name
  return 0;
} # end of SetOptionsBatch ###################################################

sub SetOptionSetGen($) ##### Set Options based on parameter (string) #########
{
  my $iRC = 0;
  my ($szOptionSet) = @_;
  if ($szOptionSet eq "compact") {
    $fPositionalTags = 0;
  }
  elsif ($szOptionSet eq "positional") {
    $fPositionalTags = 1;
  }
  elsif ($szOptionSet eq "lemmainfo") {
    $fAddInfoOut = 1;
  }
  elsif ($szOptionSet eq "nolemmainfo") {
    $fAddInfoOut = 0;
  }
  elsif ($szOptionSet eq "pdgm") {
    $fPdgmOut = 1;
  }
  elsif ($szOptionSet eq "nopdgm") {
    $fPdgmOut = 0;
  }
  elsif ($szOptionSet eq "comments") {
    $fCommentsOut = 1;
  }
  elsif ($szOptionSet eq "nocomments") {
    $fCommentsOut = 0;
  }
  elsif ($szOptionSet eq "demo") {
    SetOptionsDemoGen();
  }
  elsif ($szOptionSet eq "batch") {
    SetOptionsBatchGen();
  }
  else {
    print STDERR "$szPN: Wrong filter parameter: $szOptionSet\n";
    $iRC = 1;
  }
  return $iRC;
} # end of SetOptionSetGen ###################################################

sub SetOptionsDefaultsGen() ######## Setting runtime options (def) ###########
{
  # all options except for dictionary name should be set here
  # if need is out there for several different versions, make a copy
  # of this whole thing, rename and change options there...
  # see  also the individual overwrites (sub SetOptions...)
  my $iRC = 0;
  $iRC = SetOptionsBatchGen();
  if ($iRC != 0) { return $iRC; }
  return 0;
} # end of SetOptionsDefaultsGen #############################################

############################################################## MAIN ##########

  # Program starts here:
  $iRC = 0;
  $szPN = $0; # remember program name for messages

  # check which code to expect at input: (might be changed dynamically):
  if ($szPN =~ /iso/i) {
    $fInputCode = $ISOLATIN2;
  }
  elsif ($szPN =~ /win/i) {
    $fInputCode = $WINCP1250;
  }
  else {
    $fInputCode = $ISOLATIN2; # default for old scripts
  }

  if ($fInputCode == $WINCP1250) {
    printf STDERR 
           "$szPN: using input/output code conversion from/to Win CP1250.\n";
  }

  # check if we should run as daemon (first parameter, shift rest if yes):
  $fServ = 0;
  if (($#ARGV >= 0) &&
      ($ARGV[0] eq $szSERVER)) {
    $fServ = 1;
    shift;
    if ($#ARGV >= 0) {
      $iPort = $ARGV[0];
    }
    else {
      print STDERR "Usage: SERVER port dict [mode] [mode]\n";
      exit 985;
    }
    # untaint port:
    if ($iPort =~ /^([0-9]+)$/) {
      $iPort = $1;
    }
    else {
      print STDERR "$szPN: $iPort must be numeric.\n";
      exit 986;
    }
    if ($iPort < 30801 &&
        $iPort > 31000) {
      print STDERR "$szPN: port $iPort out of range.\n";
      exit 988;
    }
    shift; # get ready for dictionary in
  } # end of daemon check

  if ($fServ == 0) { # not as server/daemon
    # check if we should use supplied filenames instead of stdin/out:
    if (($#ARGV >= 0) &&
        ($ARGV[0] eq $szNOSTDIO)) {
      shift;
      if (($#ARGV >= 0)) {
        $szMAIn = $ARGV[0];
        shift;
        if (($#ARGV >= 0)) {
	    $szMAOut = $ARGV[0];
            shift;
        } 
        else {
          print STDERR 
            "Usage: $szPN NOSTDIO Infile Outfile [usual params]...\n";
          exit 998;
        }
      }
      else {
        print STDERR 
          "Usage: $szPN NOSTDIO Infile Outfile [usual params]...\n";
        exit 997;
      }
      close(STDIN);
      open (STDIN, $szMAIn) || do {
        print STDERR "$szPN: cannot open input file $szMAIn\n";
        exit 996;
      };
      close(STDOUT);
      $iRCCall = open STDOUT, ">$szMAOut";
      if ($iRCCall == undef) { 
        print STDERR "$szPN: cannot open output file $szMAOut\n";
        exit 995;
      };
    } # end of NOSTDIO check
  } # end of not daemon

  # log version, calling style:
  print STDERR 
      "$szPN: version $FMVERSION $FMDATE";
  if ($fServ == 1) {
    print STDERR " SERVER";
  }
  else {
    if ($szMAIn ne "" ||
        $szMAOut ne "") {
      print STDERR " NOSTDIO";
      if ($szMAIn ne "") {
        print STDERR " In: $szMAIn";
      }
      if ($szMAOut ne "") {
        print STDERR " Out: $szMAOut";
      }
    }
    else {
      print STDERR " stdin/stdout";
    }
  } # logging; not server mode
  print STDERR "\n";

  # dictionary and mode parameters:
  if (($#ARGV < 0) || ($#ARGV > 5)) {
    print STDERR 
      "Usage: $szPN Dictionary-file [Mode...] < In > Out\n";
    print STDERR 
      "(or:   $szPN NOSTDIO Filein Fileout Dictionary-file [Mode...])\n";
    print STDERR
      "      Modes:\n";
    print STDERR
      "                          positional      run w/positional tags (def)\n";
    print STDERR
      "                          compact         run w/compact tags\n";
    print STDERR
      "                          lemmainfo       output dict entry info\n";
    print STDERR
      "                          nolemmainfo     do not  - \" - (def)\n";
    print STDERR
      "                          pdgm            output paradigm in info\n";
    print STDERR
      "                          nopdgm          do not  - \" - (def)\n";
    print STDERR
      "                          comments        output comments in info\n";
    print STDERR
      "                          nocomments      do not  - \" - (def)\n";
    print STDERR
      "                                          pdgm,comments: only if lemmainfo\n";
    print STDERR
      "      Demo mode:\n";
    print STDERR
      "                          demo      for interactive testing\n";
    print STDERR
      "                                    lemmainfo,pdgm,comments on in demo\n";
    exit 999;
  }
  $DictionaryFileName = $ARGV[0]; # get dictionary name #####################

  # remember optional command line arguments:
  for ($i = 0; $i < $#ARGV; $i++) {
    $rgPar[$i] = $ARGV[$i+1];
  }

  # Initialize options: #####################################################
  $iRC = SetOptionsDefaultsGen();
  if ($iRC != 0) {
    print STDERR $szPN,
                 ": ERROR: cannot set options ",
                 "; exiting\n";
    exit 1;
  }
  # set additional options according to parameters:
  for ($i = 0; $i <= $#rgPar; $i++) {
    $iRC = SetOptionSetGen($rgPar[$i]);
    if ($iRC != 0) {
      print STDERR $szPN,
                   ": ERROR: cannot set additional options ($i)",
                   "; exiting\n";
      exit 2;
    }
  } # of loop setting options from command line

  $iRC = GetDictionaryGen($DictionaryFileName); # open and read dict in ###
  if ($iRC != 0) { # exit if any problems encountered
    print STDERR $szPN,
                 ": ERROR: cannot open and/or read dictionary ",
                 $DictionaryFileName, "; exiting\n";
    exit (10 + $iRC);
  }

  if ($fServ == 0) { # not server mode 
    $iRC = MorphGenerate(""); ###########################################
    if ($iRC != 0) { # exit if any problems encountered
      print STDERR $szPN,
              ": ERROR: cannot morphologically generate forms from (STDIN)",
              "; exiting\n";
      exit (100+$iRC);
    }
  }
  else { # server code follows
    sub REAPER {
      $waitedpid = wait;
      # loathe sysV: it makes us not only reinstate
      # the handler, but place it after the wait
      $SIG{CHLD} = \&REAPER;
    }
    sub logmsg { print STDERR "$szPN @_ at ", scalar localtime, "\n" }
    $szDataInCmd = "0f3b"; # command code for server: data follows
    $szFromWhCmd = "8b01"; # command code for server: client IP
    $szSetOptCmd = "a10a"; # command code for server: option follows
    $szWaitCmd = "5aa0"; # command code for server: wait after processing
    $szStopCmd = "02ab"; # command code for server: stop
    $szVerCmd = "e7ce"; # command code for client: dict version info
    $szEofCmd = "800d"; # command code for client: end of data
    BEGIN { $ENV{PATH} = '/usr/ucb:/bin' }
    $EOL = "\012";

    my $iProto = getprotobyname('tcp');

    socket(Server, PF_INET, SOCK_STREAM, $iProto) 
      || die "socket: $!";
    setsockopt(Server, SOL_SOCKET, SO_REUSEADDR, pack("l", 1))
      || die "setsockopt: $!";
    bind(Server, sockaddr_in($iPort, INADDR_ANY))
      || die "bind: $!";
    listen(Server,1) 
      || die "listen: $!";

    logmsg "server started on port $iPort, 1 conn.";

    my $iPaddr;

    $SIG{CHLD} = \&REAPER;

    $szWaitAfter = $iWaitEndDefault;
    # main client-listening loop:
    for ( ; $iPaddr = accept(Client,Server); close Client) {
      my ($iPort,$iAddr) = sockaddr_in($iPaddr);

#      my $szClientHostname = gethostbyaddr($iAddr,AF_INET);
#      logmsg "connection from $szClientHostname [",
#                   inet_ntoa($iAddr), "]",
#                   "at port $iPort"; # comment out once it runs well
#                                     # (hopefully only from us)

      open(STDIN,"<&Client") or die "can't dup STDIN";
      open(STDOUT,">&Client") or die "can't dup STDOUT";
      STDOUT->autoflush();
      $szIn = <STDIN>;
      chomp $szIn;
      $fDataExpected = 0;
      $szClient = "unknown";
      $szLastOpt = "default";
      if (defined($szIn)) {
        $fDataExpected = 1;
        while ($szIn !~ /^$szDataInCmd[ ]/o) {
          $cLengthIn = length($szIn);
          if ($cLengthIn > 2800) {
            logmsg "Too long input ($cLengthIn). Aborting.";
	    exit 912;
          }
          if ($szIn =~ /^$szSetOptCmd[ ]/o) {
            $szLastOpt = substr($szIn,5);
	    $iRC = SetOptionSetGen($szLastOpt);
	    if ($iRC != 0) {
	      logmsg "invalid option $szIn";
  	      $fDataExpected = 0;
	      last;
	    }
          }
          elsif ($szIn =~ /^$szFromWhCmd[ ]/o) {
            $szClient = substr($szIn,5);
          }
          elsif ($szIn =~ /^$szWaitCmd[ ]/o) {
            $szWaitAfter = substr($szIn,5);
          }
          elsif ($szIn =~ /^$szStopCmd[ ]/o) {
	      logmsg "Stop command received. Exitting.";
	    exit 990;
          }
          else { # wrong command code!
            logmsg "Unauthorized command code. Hacker's attack?";
	    exit 911;
          }
          $szIn = <STDIN>;
          if (!defined($szIn)) {
	    $fDataExpected = 0;
	    last;
          }
          chomp $szIn;
        }
      } 
      if ($fDataExpected == 1) { # yes, work on it
        if ($szClient eq "unknown") {
          logmsg "Client address unknown. Cannot serve uknown clients, you hacker.";
          sleep 5;
          exit 915;
	}
        $cLengthIn = length($szIn);
        if ($cLengthIn > 2800) {
          logmsg "Too long input ($cLengthIn). Aborting.";
          exit 913;
        }
        $szDataToAnalyze = substr($szIn,5);

        logmsg "$szDataToAnalyze (opt $fAddInfoOut/$fCommentsOut/$fPdgmOut)".
                       "$szClient";
        print STDOUT "$szVerCmd ", "Code version: ", $FMVERSION, " ", $FMDATE, 
                                   "@ Dictionary: ",
                                   "$szDictIDAndVersionInfo ",
                                   "@ ",
                                   "Endings: $cEndingCount, ",
                                   "Roots: $cRootCount",
                                   $EOL;
        $iRC = MorphGenerate($szDataToAnalyze); ###########################
        if ($iRC != 0) { # exit if any problems encountered
          print STDERR $szPN,
           ": ERROR: cannot generate from input $szDataToAnalyze\n",
        }
        print STDOUT "$szEofCmd ", $EOL;
      }
      else { # no data (just set option, maybe)
        print STDOUT "$szEofCmd ", "ERROR", $EOL;
      }
      $iWaitEnd = $szWaitAfter;
      sleep $iWaitEnd; 
    } # end of server listening code loop
  } # end of server code

  0;
######################################################### END OF MAIN ########
