package HB_CharCode8;
require 5.004;
use strict;

my $VERSION = '1.1.0'; # Addde unient (ISO 10646, using numeric ent names)
#my $VERSION = '1.0.0'; # first six (well, five :) codetables

=head1 NAME

HB_CharCode8 - code/decode "characters" to/from 8-bit code

=head1 SYNOPSIS

    HB_CharCode8SetTable("$szCodingTableName>");
    $szNamedCharString = HB_CharCode8ToString("$sz8bitCodedString");
    $sz8bitCodedString = HB_CharCode8FromString("$szNamedCharString");
    $chNamed = HB_CharCode8ToChar("$ch8bit");
    $ch8bit = HB_CharCode8FromChar("$chNamed");
    @AvailableCodes = HB_CharCode8ListCodes();

=head1 DESCRIPTION

The C<HB_CharCode8SetTable> sets a coding table for subsequent coding
by the other four subroutines. Allowable names are C<windows-1250>,
C<iso-8859-2>, C<tex>, C<ent>, C<pseudo>, and C<unaccented>. The 8-bit
code used in these tables is ISO Latin-2 for those tables, but
future tables might facilitate translation into other code tables.

=head1 BUGS

Only IRV-based code expected (tables deal with >= 128 codes only).
(I.e., no EBCDIC etc. conversion possible.)

No error reports, even when code not found or duplicate mapping in
either direction. Output chosen deterministically but no particular
algorithm is used.

=head1 AUTHOR

Jan Hajic, hajic@ufal.mff.cuni.cz

=cut


use vars qw(@ISA @EXPORT @EXPORT_OK
       $ActiveTableTo $ActiveTableFrom
       @ToUnaccented @ToPseudo @ToTeX @ToEnt @ToIsoL2 @ToWin1250
       @ToIsoNumEnt
       @FromUnaccented %FromPseudo %FromTeX %FromEnt @FromIsoL2 @FromWin1250
       %FromIsoNumEnt
  );

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(
  HB_CharCode8SetTable 
  HB_CharCode8ToChar HB_CharCode8FromChar 
  HB_CharCode8ToString HB_CharCode8FromString
  HB_CharCode8ListCodes
 );
@EXPORT_OK = qw();


# tables: #################################################################

# table names: NB Don't forget add new names to HB_CharCode8ListCodes!!!!
my $szTabNameIsoL2 = "iso-8859-2";
my $szTabNameWin1250 = "windows-1250";
my $szTabNameTeX = "tex";
my $szTabNameEnt = "ent";
my $szTabNamePseudo = "pseudo";
my $szTabNameUnaccented = "unaccented";
my $szTabNameUniEnt = "unient";

# ISO 10646 table, using numeric character entity names
@ToIsoNumEnt = (
"?","?","?","?","?","?","?","?", "?","?","?","?","?","?","?","?", 
"?","?","?","?","?","?","?","?", "?","?","?","?","?","?","?","?", 
"&#160;","&#260;","&#728;","&#321;","&#164;","&#317;","&#346;","&#167;",
"&#168;","&#352;","&#350;","&#356;","&#377;","&#150;","&#381;","&#379;",
"&#730;","&#261;","&#731;","&#322;","&#714;","&#318;","&#347;","&#711;",
"&#184;","&#353;","&#351;","&#357;","&#378;","&#733;","&#382;","&#380;",
"&#340;","&#193;","&#194;","&#258;","&#196;","&#313;","&#262;","&#199;",
"&#268;","&#201;","&#280;","&#203;","&#282;","&#205;","&#206;","&#270;",
"&#272;","&#323;","&#327;","&#211;","&#212;","&#336;","&#214;","&#215;",
"&#344;","&#366;","&#218;","&#368;","&#220;","&#221;","&#354;","&#223;",
"&#341;","&#225;","&#226;","&#259;","&#228;","&#314;","&#263;","&#231;",
"&#269;","&#233;","&#281;","&#235;","&#283;","&#237;","&#238;","&#271;",
"&#273;","&#324;","&#328;","&#243;","&#244;","&#337;","&#246;","&#247;",
"&#345;","&#367;","&#250;","&#369;","&#252;","&#253;","&#355;","&#729;"
);

# Unaccented code table (SetTable: unaccented)
@ToUnaccented = ( # strip accents; cannot be covnerted back
"?","?","?","?","?","?","?","?", "?","?","?","?","?","?","?","?", 
"?","?","?","?","?","?","?","?", "?","?","?","?","?","?","?","?", 
" ","A","_","L",
'$',"L","S","~",
":","S","S","T",
"Z","-","Z","Z",
"*","a",",","l",
"\'","l","s","\"",
",","s","s","t",
"z",";","z","z",
"R","A","A","A",
"A","L","C","C",
"C","E","E","E",
"E","I","I","D",
"D","N","N","O",
"O","O","O","x",
"R","U","U","U",
"U","Y","T","s",
"r","a","a","a",
"a","l","c","c",
"c","e","e","e",
"e","i","i","d",
"d","n","n","o",
"o","o","o","/",
"r","u","u","u",
"u","y","t","."
);

# Pseudo-accented code table (SetTable: Pseudo)
@ToPseudo = ( # Pseudocode, ' after: long, " after: hachek/ring, many undef.
              # backwards!
"?","?","?","?","?","?","?","?", "?","?","?","?","?","?","?","?", 
"?","?","?","?","?","?","?","?", "?","?","?","?","?","?","?","?", 
" ","A_","_","L/",
'$',"L\"","S\'","P".'$',
":","S\"","S,","T\"",
"Z\'","-","Z\"","Z.",
"*","a,",",","l/",
"\'","l\"","s\'","\"",
",","s\"","s,","t\"",
"z\'",";","z\"","z.",
"R\'","A\'","A^","A_",
"A:","L\'","C\'","C,",
"C\"","E\'","E,","E:",
"E\"","I\'","I^","D\"",
"D/","N\'","N\"","O\'",
"O^","O\;","O:","x",
"R\"","U\"","U\'","U;",
"U:","Y\'","T,","s".'$',
"r\'","a\'","a^","a_",
"a:","l\'","c\'","c,",
"c\"","e\'","e,","e:",
"e\"","i\'","i^","d\"",
"d/","n\'","n\"","o\'",
"o^","o;","o:","/",
"r\"","u\"","u\'","u;",
"u:","y\'","t,","."
);

# (La)TeX-like sequences
@ToTeX = ( # general format: \<acc>{letter}
"?","?","?","?","?","?","?","?", "?","?","?","?","?","?","?","?", 
"?","?","?","?","?","?","?","?", "?","?","?","?","?","?","?","?", 
"~","{\\c{A}}","{\\u{}}","{\\L}",
"{\\" . '$' . "{}}","{\\v{L}}","{\\\'{S}}","{\\S}",
"{\\\"{}}","{\\v{S}}","{\\c{S}}","{\\v{T}}",
"{\\\'{Z}}","\\-","{\\v{Z}}","{\\.{Z}}",
"{\\accent\'27{}}","{\\c{a}}","{\\c{}}","{\\l}",
"{\\\'{}}","{\\v{l}}","{\\\'{s}}","{\\v{}}",
"{\\c{}}","{\\v{s}}","{\\c{s}}","{\\v{t}}",
"{\\\'{z}}","{\\H{}}","{\\v{z}}","{\\.{z}}",
"{\\\'{R}}","{\\\'{A}}","{\\^{A}}","{\\u{A}}",
"{\\\"{A}}","{\\\'{L}}","{\\\'{C}}","{\\c{C}}",
"{\\v{C}}","{\\\'{E}}","{\\c{E}}","{\\\"{E}}",
"{\\v{E}}","{\\\'{I}}","{\\^{I}}","{\\v{D}}",
"{\\D}","{\\\'{N}}","{\\v{N}}","{\\\'{O}}",
"{\\^{O}}","{\\H{O}}","{\\\"{O}}","{\\times}",
"{\\v{R}}","{\\accent\'27U}","{\\\'{U}}","{\\H{U}}",
"{\\\"{U}}","{\\\'{Y}}","{\\c{T}}","{\\ss}",
"{\\\'{r}}","{\\\'{a}}","{\\^{a}}","{\\u{a}}",
"{\\\"{a}}","{\\\'{l}}","{\\\'{c}}","{\\c{c}}",
"{\\v{c}}","{\\\'{e}}","{\\c{e}}","{\\\"{e}}",
"{\\v{e}}","{\\\'{\\i}}","{\\^{\\i}}","{\\v{d}}",
"{\\d}","{\\\'{n}}","{\\v{n}}","{\\\'{o}}",
"{\\^{o}}","{\\H{o}}","{\\\"{o}}","{\\div}",
"{\\v{r}}","{\\accent\'27u}","{\\\'{u}}","{\\H{u}}",
"{\\\"{u}}","{\\\'{y}}","{\\c{t}}","{\\.{}}"
);

# Entity name table (SetTable: ent)
@ToEnt = ( # SGML entity names, without leading/ending &;
"?","?","?","?","?","?","?","?", "?","?","?","?","?","?","?","?", 
"?","?","?","?","?","?","?","?", "?","?","?","?","?","?","?","?", 
"&nbsp;","&Aogon;","&breve;","&Lstrok;",
"&curren;","&Lcaron;","&Sacute;","&sect;",
"&uml;","&Scaron;","&Scedil;","&Tcaron;",
"&Zacute;","&shy;","&Zcaron;","&Zdot;",
"&deg;","&aogon;","&ogon;","&lstrok;",
"&acute;","&lcaron;","&sacute;","&caron;",
"&cedil;","&scaron;","&scedil;","&tcaron;",
"&zacute;","&dblac;","&zcaron;","&zdot;",
"&Racute;","&Aacute;","&Acirc;","&Abreve;",
"&Auml;","&Lacute;","&Cacute;","&Ccedil;",
"&Ccaron;","&Eacute;","&Eogon;","&Euml;",
"&Ecaron;","&Iacute;","&Icirc;","&Dcaron;",
"&Dstrok;","&Nacute;","&Ncaron;","&Oacute;",
"&Ocirc;","&Odblac;","&Ouml;","&times;",
"&Rcaron;","&Uring;","&Uacute;","&Udblac;",
"&Uuml;","&Yacute;","&Tcedil;","&szlig;",
"&racute;","&aacute;","&acirc;","&abreve;",
"&auml;","&lacute;","&cacute;","&ccedil;",
"&ccaron;","&eacute;","&eogon;","&euml;",
"&ecaron;","&iacute;","&icirc;","&dcaron;",
"&dstrok;","&nacute;","&ncaron;","&oacute;",
"&ocirc;","&odblac;","&ouml;","&divide;",
"&rcaron;","&uring;","&uacute;","&udblac;",
"&uuml;","&yacute;","&tcedil;","&dot;"
);

@ToWin1250 = ( # just slightly different... (NB: for inverse,
               # the later code gets precedence
               # because of the way the reverse table is initialized)
"\200","\201","\202","\203","\204","\205","\206","\207",
"\210","\211","\212","\213","\214","\215","\216","\217",
"\220","\221","\222","\223","\224","\225","\226","\227",
"\230","\231","\232","\233","\234","\235","\236","\237", 
"","","","","","","\214","", 
  "","\212","","\215","\217","","\216","",
"","","","","","","\234","", 
  "","\232","","\235","\237","","\236","",
"","","","","","","","", "","","","","","","","",
"","","","","","","","", "","","","","","","","",
"","","","","","","","", "","","","","","","","",
"","","","","","","","", "","","","","","","",""
);

@ToIsoL2 = ( # identity, here; just for completeness & reference
"?","?","?","?","?","?","?","?", "?","?","?","?","?","?","?","?", 
"?","?","?","?","?","?","?","?", "?","?","?","?","?","?","?","?", 
"","","","","","","","", "","","","","","","","",
"","","","","","","","", "","","","","","","","",
"","","","","","","","", "","","","","","","","",
"","","","","","","","", "","","","","","","","",
"","","","","","","","", "","","","","","","","",
"","","","","","","","", "","","","","","","",""
);

@FromIsoL2 = ( # identity, here; just for completeness & reference
"?","?","?","?","?","?","?","?", "?","?","?","?","?","?","?","?", 
"?","?","?","?","?","?","?","?", "?","?","?","?","?","?","?","?", 
"","","","","","","","", "","","","","","","","",
"","","","","","","","", "","","","","","","","",
"","","","","","","","", "","","","","","","","",
"","","","","","","","", "","","","","","","","",
"","","","","","","","", "","","","","","","","",
"","","","","","","","", "","","","","","","",""
);

my $DefFromUnaccented = 1; # already defined statically (use IsoL2)
my $DefFromPseudo = 0; # must be defined dynamically
my $DefFromEnt = 0; # must be defined dynamically
my $DefFromTeX = 0; # must be defined dynamically
my $DefFromWin1250 = 0; # must be defined dynamically
my $DefFromIsoL2 = 1; # already defined statically
my $DefFromUniEnt = 0; # must be defined dynamically

my (%FromPseudo, %FromTeX, %FromEnt, 
    @FromUnaccented, @FromWin1250, @FromIsoL2); # inverse tables or hashes

my $ActiveTableTo = \@ToIsoL2; # reference to active table; default: id
my $ActiveTableFrom = \@FromIsoL2; # reference to active table: default id:

sub HB_CharCode8ListCodes()
{
  my @CodeList;
  @CodeList = (
    $szTabNameIsoL2,
    $szTabNameWin1250,
    $szTabNameTeX,
    $szTabNameEnt,
    $szTabNamePseudo,
    $szTabNameUnaccented,
    $szTabNameUniEnt,
  );
  return @CodeList;
} # end of HB_CharCode8ListCodes()

sub InitFromTable($)
{
  # initalize inverse ("from") table; hash needed for string type of code
  my ($szCodeName) = @_;
  my $i;
  my $iToIndex;
  my $ActiveTable; 

  # print STDERR "Initializing \"from\" table for $szCodeName\n"; # dbg
  if ($szCodeName eq $szTabNamePseudo) {
    $ActiveTable = \@ToPseudo;
    for ($i = 0; $i < 128; $i++) {
      $FromPseudo{$ActiveTable->[$i]} = chr($i+128);
    }
    $DefFromPseudo = 1;
  }
  elsif ($szCodeName eq $szTabNameTeX) {
    $ActiveTable = \@ToTeX;
    for ($i = 0; $i < 128; $i++) {
      $FromTeX{$ActiveTable->[$i]} = chr($i+128);
    }
    $DefFromTeX = 1;
  }
  elsif ($szCodeName eq $szTabNameEnt) {
    $ActiveTable = \@ToEnt;
    for ($i = 0; $i < 128; $i++) {
      $FromEnt{$ActiveTable->[$i]} = chr($i+128);
    }
    $DefFromEnt = 1;
  }
  elsif ($szCodeName eq $szTabNameUniEnt) {
    $ActiveTable = \@ToIsoNumEnt;
    for ($i = 0; $i < 128; $i++) {
      $FromIsoNumEnt{$ActiveTable->[$i]} = chr($i+128);
    }
    $DefFromUniEnt = 1;
  }
  elsif ($szCodeName eq $szTabNameWin1250) {
    $ActiveTable = \@ToWin1250;
    for ($i = 0; $i < 128; $i++) {
      $FromWin1250[$i] = "?"; # make sure every element gets a value
    }
    for ($i = 0; $i < 128; $i++) {
      $iToIndex = ord($ActiveTable->[$i])-128;
      if ($iToIndex >= 0 && $iToIndex < 128) { # if > 128 code char, go ahead
        $FromWin1250[$iToIndex] = chr($i+128);
      }
    }
    $DefFromWin1250 = 1;
  }
  elsif ($szCodeName eq $szTabNameIsoL2) {
    # do nothing - ID table initialized "statically"
  }
  elsif ($szCodeName eq $szTabNameUnaccented) {
    # do nothing - use ID table, initialized "statically"
  }
  else {
    return 1; # wrong/unimplemented table
  }
  return 0;
} # end of InitFromTable

sub HB_CharCode8SetTable($)
{
  my ($szCodeName) = @_;
  my $iRC;
  if ($szCodeName eq $szTabNamePseudo) {
    $ActiveTableTo = \@ToPseudo;
    if ($DefFromPseudo == 0) {
      $iRC = InitFromTable($szCodeName);
      if ($iRC != 0) { return 10+$iRC; }
    }
    $ActiveTableFrom = \%FromPseudo;
  }
  elsif ($szCodeName eq $szTabNameUniEnt) {
    $ActiveTableTo = \@ToIsoNumEnt;
    if ($DefFromUniEnt == 0) {
      $iRC = InitFromTable($szCodeName);
      if ($iRC != 0) { return 10+$iRC; }
    }
    $ActiveTableFrom = \%FromIsoNumEnt;
  }
  elsif ($szCodeName eq $szTabNameEnt) {
    $ActiveTableTo = \@ToEnt;
    if ($DefFromEnt == 0) {
      $iRC = InitFromTable($szCodeName);
      if ($iRC != 0) { return 10+$iRC; }
    }
    $ActiveTableFrom = \%FromEnt;
  }
  elsif ($szCodeName eq $szTabNameTeX) {
    $ActiveTableTo = \@ToTeX;
    if ($DefFromTeX == 0) {
      $iRC = InitFromTable($szCodeName);
      if ($iRC != 0) { return 10+$iRC; }
    }
    $ActiveTableFrom = \%FromTeX;
  }
  elsif ($szCodeName eq $szTabNameWin1250) {
    $ActiveTableTo = \@ToWin1250;
    if ($DefFromWin1250 == 0) {
      $iRC = InitFromTable($szCodeName);
      if ($iRC != 0) { return 10+$iRC; }
    }
    $ActiveTableFrom = \@FromWin1250;
  }
  elsif ($szCodeName eq $szTabNameUnaccented) {
    $ActiveTableTo = \@ToUnaccented;
    if ($DefFromUnaccented == 0) {
      return 17; # should always be defined - init'ed statically
    }
    $ActiveTableFrom = \@FromIsoL2; # use ID table
  }
  elsif ($szCodeName eq $szTabNameIsoL2) {
    $ActiveTableTo = \@ToIsoL2;
    if ($DefFromIsoL2 == 0) {
      return 19; # should always be defined - init'ed statically
    }
    $ActiveTableFrom = \@FromIsoL2;
  }
  else {
    return 18; # wrong/unimplemnted table
  }
  return 0;
} # end of HB_CharCode8SetTable

sub HB_CharCode8ToChar($)
{
  my $iTable; # set to the appropriate index in table
  my ($chIn) = @_;
  $iTable = ord($chIn);
  if ($iTable >= 128) {
    return ($ActiveTableTo->[$iTable-128]);
  }
  else {
    return $chIn;
  }
} # end of HB_CharCode8ToChar

sub HB_CharCode8FromChar($)
{
  my $iTable; # set to the appropriate index in table
  my ($chIn) = @_;
  my $chOut; # result
  if (ref($ActiveTableFrom) eq "ARRAY") { # simple array indexing
                                          # possible
    $iTable = ord($chIn);
    if ($iTable >= 128) {
      $chOut = ($ActiveTableFrom->[$iTable-128]);
    }
    else {
      $chOut = $chIn;
    }
  }
  elsif (ref($ActiveTableFrom) eq "HASH") { # using hash string --> char
    if (exists($ActiveTableFrom->{$chIn}) &&
        defined($ActiveTableFrom->{$chIn})) {
      $chOut = ($ActiveTableFrom->{$chIn});
    }
    else {
      $chOut = $chIn;
    }
  }
  else {
    return "?"; # fatal internal error - wrong table type
  }
  return $chOut;
} # end of HB_CharCode8FromChar

sub HB_CharCode8ToString($)
{
  my ($szStringIn) = @_;
  my $szStringOut = "";
  my @szIn = split //, $szStringIn;
  my $cIn = @szIn; # length (size) of input
  my $i;

  for ($i = 0; $i < $cIn; $i++) {
    $szStringOut .= HB_CharCode8ToChar($szIn[$i]); 
  }
  return $szStringOut;
} # end of HB_CharCode8ToString

sub HB_CharCode8FromString($)
{
  my ($szStringIn) = @_;
  my $szStringOut = "";
  my @szIn = split //, $szStringIn;
  my $cIn = @szIn; # length (size) of input
  my $chIn1; # for complicated string-to-char conversions
  my $chIn; # for complicated string-to-char conversions
  my $szMatch; # for complicated string-to-char conversions (tex)
  my $i;
  my $iNext;

  if (ref($ActiveTableFrom) eq "ARRAY") { # simple array indexing
                                          # possible
    for ($i = 0; $i < $cIn; $i++) {
      $szStringOut .= HB_CharCode8FromChar($szIn[$i]); 
    }
  }
  elsif (ref($ActiveTableFrom) eq "HASH") { # using hash string --> char
    if ($ActiveTableFrom == \%FromPseudo) {
      for ($i = 0; $i < $cIn; $i++) {
        $chIn = $szIn[$i];
        $iNext = $i + 1;
        if ($iNext < $cIn &&
            $szIn[$iNext] =~ /['"]/) { # yes, pseudo accented char
          $chIn .= $szIn[$iNext];
          $i++;
          $szStringOut .= HB_CharCode8FromChar($chIn); # decode pseudocode
        }
        else {
          $szStringOut .= $chIn;  # just copy character 
        }
      } # end of main for loop over string
    } # end of pseudocode decoding
    elsif ($ActiveTableFrom == \%FromEnt) {
      for ($i = 0; $i < $cIn; $i++) {
        $chIn = $szIn[$i];
        if ($chIn eq "&") { # possibly an entity
          $i++;
          while ($i < $cIn) {
            $chIn1 = $szIn[$i];
            if ($chIn1 eq ";") {
              $chIn .= $chIn1; # even the final ";"
              last; # whole entity name extracted into $chIn
            }
            elsif ($chIn1 =~ /^[_0-9a-zA-Z]$/) {
              $chIn .= $chIn1; # add a valid char
              $i++;
            }
            else {
              $i--; # go back so that current char read again
                    # NB cannot go < 0 since we started at 1 at least
              last;
            }
          } # end of extracting entity name into $chIn
          if ($chIn1 eq ";") {
            $szStringOut .= HB_CharCode8FromChar($chIn); 
          }
          else {
            $szStringOut .= $chIn; 
          }
        } # end of initial-"&" processing
        else {
          $szStringOut .= $chIn;
        } # of non-"&" processing
      } # end of loop ove input string
    } # end of "ent" as active table
    elsif ($ActiveTableFrom == \%FromIsoNumEnt) {
      for ($i = 0; $i < $cIn; $i++) {
        $chIn = $szIn[$i];
        if ($chIn eq "&") { # possibly an entity
          $i++;
          while ($i < $cIn) {
            $chIn1 = $szIn[$i];
            if ($chIn1 eq ";") {
              $chIn .= $chIn1; # even the final ";"
              last; # whole entity name extracted into $chIn
            }
            elsif ($chIn1 =~ /^[\#0-9]$/) { # should check better but...
              $chIn .= $chIn1; # add a valid char
              $i++;
            }
            else {
              $i--; # go back so that current char read again
                    # NB cannot go < 0 since we started at 1 at least
              last;
            }
          } # end of extracting entity name into $chIn
          if ($chIn1 eq ";") {
            $szStringOut .= HB_CharCode8FromChar($chIn); 
          }
          else {
            $szStringOut .= $chIn; 
          }
        } # end of initial-"&" processing
        else {
          $szStringOut .= $chIn;
        } # of non-"&" processing
      } # end of loop ove input string
    } # end of "unient" as active table
    elsif ($ActiveTableFrom == \%FromTeX) {
      while ($szStringIn ne "") {
        if ($szStringIn =~ 
              /^([{](((\\['`^"~=.rkuvHcdb])([ \t\r\n])?[{](([a-zA-Z])|(\\i))[}])|(\\(([lLdDS])|(ss))))[}])/) {
          # exact match with preferred string
          $szMatch = $1; 
          $szStringIn = substr($szStringIn,length($szMatch));
          $chIn = $szMatch;
          if (substr($szMatch,3,1) =~ /[ \t\r\n]/) {
            $chIn = substr($szMatch,0,3) . 
                    substr($szMatch,4);
          }
          $szStringOut .= HB_CharCode8FromChar($chIn); 
        }
        elsif ($szStringIn =~ 
             /^((\\['`^"~=.rkuvHcdb])[{](([a-zA-Z])|(\\i))[}])/) {
          # same as above (no spec chars) but no enclosing curly braces,
          # regardless what follows
          $szMatch = $1; 
          $szStringIn = substr($szStringIn,length($szMatch));
          $chIn = "{" . $szMatch . "}"; 
          $szStringOut .= HB_CharCode8FromChar($chIn); 
        }
        elsif ($szStringIn =~
            /^([{]\\accent[']27([ \t\r\n])?(([a-zA-Z]))[}])/) {
          # accent ring with no curly brackets around letter, not \i
          $szMatch = $1; 
          $szStringIn = substr($szStringIn,length($szMatch));
          $chIn = "{\\accent\'27" . 
                       substr($szMatch,length($szMatch)-2,1) . "}";
          $szStringOut .= HB_CharCode8FromChar($chIn); 
        }
        elsif ($szStringIn =~
            /^([{]\\accent[']23([ \t\r\n])?(([a-zA-Z]))[}])/) {
          # accent acute with no curly brackets around letter, not \i
          $szMatch = $1; 
          $szStringIn = substr($szStringIn,length($szMatch));
          $chIn = "{\\'{" . substr($szMatch,length($szMatch)-2,1) . "}}";
          $szStringOut .= HB_CharCode8FromChar($chIn); 
        }
        elsif ($szStringIn =~
            /^([{]\\accent[']23([ \t\r\n])?((\\[i]))[}])/) {
          # accent acute with no curly brackets around \i
          $szMatch = $1; 
          $szStringIn = substr($szStringIn,length($szMatch));
          $chIn = "{\\'{" . substr($szMatch,length($szMatch)-3,2) . "}}";
          $szStringOut .= HB_CharCode8FromChar($chIn); 
        }
        elsif ($szStringIn =~ /^([{]\\['`^"~=.]([ \t\r\n])?[a-zA-Z][}])/) {
          # nonletter accents, no curly brackets around letter
          # regardless what follows
          $szMatch = $1; 
          $szStringIn = substr($szStringIn,length($szMatch));
          $chIn = substr($szMatch,0,3) . "{" . 
                  substr($szMatch,length($szMatch)-2,2) . "}";
          $szStringOut .= HB_CharCode8FromChar($chIn); 
        }
        elsif ($szStringIn =~ /^([{]\\['`^"~=.]([ \t\r\n])?\\[i][}])/) {
          # nonletter accents, no curly brackets around \i
          # regardless what follows
          $szMatch = $1; 
          $szStringIn = substr($szStringIn,length($szMatch));
          $chIn = substr($szMatch,0,3) . "{" . 
                  substr($szMatch,length($szMatch)-3,3) . "}";
          $szStringOut .= HB_CharCode8FromChar($chIn); 
        }
        elsif ($szStringIn =~ /^([{]\\['`^"~=.rkuvHcdb][ \t\r\n][a-zA-Z][}])/) {
          # space sep., no curly brackets around letter, not \i
          # regardless what follows
          $szMatch = $1; 
          $szStringIn = substr($szStringIn,length($szMatch));
          $chIn = substr($szMatch,0,3) . "{" . 
                  substr($szMatch,4) . "}";
          $szStringOut .= HB_CharCode8FromChar($chIn); 
        }
        elsif ($szStringIn =~ /^(\\['`^"~=.rkuvHcdb][ \t\r\n][a-zA-Z])/) {
          # no braces, space sep., no curly brackets around letter, not \i
          # regardless what follows
          $szMatch = $1; 
          $szStringIn = substr($szStringIn,length($szMatch));
          $chIn = "{" . substr($szMatch,0,2) . "{" . 
                        substr($szMatch,3) . "}}";
          $szStringOut .= HB_CharCode8FromChar($chIn); 
        }
        elsif ($szStringIn =~ /^(\\['`^"~=.rkuvHcdb][ \t\r\n]\\[i][ \t\n\r])/) {
          # no braces, space sep., no curly brackets around letter, \i
          # followed by nonletter
          $szMatch = $1; 
          $szStringIn = substr($szStringIn,length($szMatch));
          $chIn = "{" . substr($szMatch,0,2) . "{" . 
                        substr($szMatch,3,2) . "}}";
          $szStringOut .= HB_CharCode8FromChar($chIn); 
        }
        elsif ($szStringIn =~ /^(\\['`^"~=.rkuvHcdb][ \t\r\n]\\[i])[^a-zA-Z]/) {
          # no braces, space sep., no curly brackets around letter, \i
          # followed by nonletter
          $szMatch = $1; 
          $szStringIn = substr($szStringIn,length($szMatch));
          $chIn = "{" . substr($szMatch,0,2) . "{" . 
                        substr($szMatch,3) . "}}";
          $szStringOut .= HB_CharCode8FromChar($chIn); 
        }
        elsif ($szStringIn =~ /^(\\['`^"~=.rkuvHcdb][ \t\r\n][{][a-zA-Z][}])/) {
          # no braces, space sep., curly brackets around letter, not \i
          # regardless what follows
          $szMatch = $1; 
          $szStringIn = substr($szStringIn,length($szMatch));
          $chIn = "{" . substr($szMatch,0,2) . 
                        substr($szMatch,3) . "}";
          $szStringOut .= HB_CharCode8FromChar($chIn); 
        }
        elsif ($szStringIn =~ /^(\\['`^"~=.rkuvHcdb][ \t\r\n][{]\\[i][}])/) {
          # no braces, space sep., curly brackets around letter, \i
          # regardless what follows
          $szMatch = $1; 
          $szStringIn = substr($szStringIn,length($szMatch));
          $chIn = "{" . substr($szMatch,0,2) .
                        substr($szMatch,3) . "}";
          $szStringOut .= HB_CharCode8FromChar($chIn); 
        }
        else {
          $chIn = substr($szStringIn,0,1);
          $szStringIn = substr($szStringIn,length($chIn));
          $szStringOut .= $chIn;
        } # of else: no tex-like char found
      } # end of loop ove input string
    }
  }
  else {
    return "?" . $szStringIn; # fatal internal error - wrong table type
  }
  return $szStringOut;
} # end of HB_CharCode8FromString

1;
